
// SLLinxWatcherDlg.cpp : implementation file
//

#include "stdafx.h"
#include "SLLinxWatcher.h"
#include "SLLinxWatcherDlg.h"
#include "EditInfoDlg.h"
#include "afxdialogex.h"
#include <algorithm>
#include <iterator>
#include <vector>
#include <atlpath.h>
#include <utility>


#ifdef _DEBUG
#define new DEBUG_NEW
#endif

// CSLLinxWatcherDlg dialog

CSLLinxWatcherDlg::CSLLinxWatcherDlg(CWnd* pParent /*=NULL*/)
	: CDialogEx(IDD_SLLINXWATCHER_DIALOG, pParent)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

// save/restore global settings
static const TCHAR *szCaption = _T("SLLinxWatcher");
static const TCHAR *szSection = _T("");
static const TCHAR *szSrcLynxEvtPath = _T("SourceLynxEvt_Path");
static const TCHAR *szPlayersPath = _T("PlayersCSV_Path");
static const TCHAR *szSrcResultCsvPath = _T("SourceResultCSV_Path");
static const TCHAR *szDstStartCsvPath = _T("TargetStartCSV_Path");
static const TCHAR *szDstFinishCsvPath = _T("TargetFinishCSV_Path");
static const TCHAR *szProjectPath = _T("Project_Path");
//
void CSLLinxWatcherDlg::SaveSettings(CWinApp *pApp)
{
	pApp->WriteProfileString(szSection, szSrcLynxEvtPath, m_szSrcLynxEvtPath);
	pApp->WriteProfileString(szSection, szPlayersPath, m_szPlayersPath);
	pApp->WriteProfileString(szSection, szSrcResultCsvPath, m_szSrcResultCsvPath);
	pApp->WriteProfileString(szSection, szDstStartCsvPath, m_szDstStartCsvPath);
	pApp->WriteProfileString(szSection, szDstFinishCsvPath, m_szDstFinishCsvPath);
	pApp->WriteProfileString(szSection, szProjectPath, m_szProjectPath);
}
//
void CSLLinxWatcherDlg::LoadSettings(CWinApp *pApp)
{
	m_szSrcLynxEvtPath = pApp->GetProfileString(szSection, szSrcLynxEvtPath, m_szSrcLynxEvtPath);
	m_szPlayersPath = pApp->GetProfileString(szSection, szPlayersPath, m_szPlayersPath);
	m_szSrcResultCsvPath = pApp->GetProfileString(szSection, szSrcResultCsvPath, m_szSrcResultCsvPath);
	m_szDstStartCsvPath = pApp->GetProfileString(szSection, szDstStartCsvPath, m_szDstStartCsvPath);
	m_szDstFinishCsvPath = pApp->GetProfileString(szSection, szDstFinishCsvPath, m_szDstFinishCsvPath);
	m_szProjectPath = pApp->GetProfileString(szSection, szProjectPath, m_szProjectPath);
}

void CSLLinxWatcherDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialogEx::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_COMPETITION, m_Competion);
	DDX_Control(pDX, IDC_RACE, m_RaceList);
	DDX_Control(pDX, IDC_WATCH_RESULT, m_WatchResult);
	DDX_Control(pDX, IDC_PLAYER_LIST, m_PlayerList);
}

BEGIN_MESSAGE_MAP(CSLLinxWatcherDlg, CDialogEx)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_BN_CLICKED(IDC_IMPORT, &CSLLinxWatcherDlg::OnBnClickedImport)
	ON_BN_CLICKED(IDC_WATCH_PATH, &CSLLinxWatcherDlg::OnBnClickedWatchPath)
	ON_BN_CLICKED(IDC_START_CSV_PATH, &CSLLinxWatcherDlg::OnBnClickedStartCsvPath)
	ON_BN_CLICKED(IDC_FINISH_CSV_PATH, &CSLLinxWatcherDlg::OnBnClickedFinishCsvPath)
	ON_BN_CLICKED(IDC_SAVE_ALL, &CSLLinxWatcherDlg::OnBnClickedSaveAll)
	ON_BN_CLICKED(IDC_LOAD_ALL, &CSLLinxWatcherDlg::OnBnClickedLoadAll)
	ON_CBN_SELCHANGE(IDC_RACE, &CSLLinxWatcherDlg::OnCbnSelchangeRace)
	ON_BN_CLICKED(IDC_SAVE_RACE, &CSLLinxWatcherDlg::OnBnClickedSaveRace)
	ON_BN_CLICKED(IDC_SAVE_FINISH, &CSLLinxWatcherDlg::OnBnClickedSaveFinish)
	ON_BN_CLICKED(IDC_WATCH_RESULT, &CSLLinxWatcherDlg::OnBnClickedWatchResult)
	ON_WM_TIMER()
	ON_NOTIFY(NM_DBLCLK, IDC_PLAYER_LIST, &CSLLinxWatcherDlg::OnNMDblclkPlayerList)
	ON_BN_CLICKED(IDC_EXIT, &CSLLinxWatcherDlg::OnBnClickedExit)
	ON_BN_CLICKED(IDC_DB2, &CSLLinxWatcherDlg::OnBnClickedDb2)
END_MESSAGE_MAP()


// CSLLinxWatcherDlg message handlers

BOOL CSLLinxWatcherDlg::OnInitDialog()
{
	CDialogEx::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		BOOL bNameValid;
		CString strAboutMenu;
		bNameValid = strAboutMenu.LoadString(IDS_ABOUTBOX);
		ASSERT(bNameValid);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	if (!LoadCountryMap()) {
		MessageBox(_T("There is no file with country ID!\r\n\r\nPut file \'country.csv\' (ANSI) near EXE-file."), szCaption, MB_ICONSTOP);
		EndDialog(IDCANCEL);
	}

	// Init locals
	RECT rc;  int cx[eLastColumn];
	m_PlayerList.GetClientRect(&rc);
	rc.right -= rc.left;
	cx[eTrackColumn] = rc.right / 25;
	cx[eRankColumn] = rc.right / 12;
	cx[eBipColumn] = rc.right / 13;
	cx[eCountryColumn] = rc.right / 11;
	cx[eTimeColumn] = rc.right / 10;
	rc.right -= cx[eTrackColumn] + cx[eRankColumn] + cx[eBipColumn] + cx[eTimeColumn] + cx[eCountryColumn];
	cx[eFamilyColumn] = rc.right / 2;
	cx[eNameColumn] = rc.right - cx[eFamilyColumn];
	m_PlayerList.SetExtendedStyle(LVS_EX_GRIDLINES | LVS_EX_FULLROWSELECT);
	m_PlayerList.InsertColumn(eTrackColumn, _T("#"), LVCFMT_CENTER, cx[eTrackColumn]);
	m_PlayerList.InsertColumn(eRankColumn, _T("Rank"), LVCFMT_CENTER, cx[eRankColumn]);
	m_PlayerList.InsertColumn(eBipColumn, _T("BIP"), LVCFMT_CENTER, cx[eBipColumn]);
	m_PlayerList.InsertColumn(eFamilyColumn, _T("Family"), LVCFMT_LEFT, cx[eFamilyColumn]);
	m_PlayerList.InsertColumn(eNameColumn, _T("Name"), LVCFMT_LEFT, cx[eNameColumn]);
	m_PlayerList.InsertColumn(eTimeColumn, _T("Time"), LVCFMT_LEFT, cx[eTimeColumn]);
	m_PlayerList.InsertColumn(eCountryColumn, _T("Country"), LVCFMT_CENTER, cx[eCountryColumn]);

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CSLLinxWatcherDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CDialogEx dlgAbout(IDD_ABOUTBOX);
		dlgAbout.DoModal();
	}
	else
	{
		CDialogEx::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CSLLinxWatcherDlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialogEx::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CSLLinxWatcherDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

void CSLLinxWatcherDlg::OnBnClickedImport()
{
	CFileDialog dlg(TRUE, _T(".evt"), m_szSrcLynxEvtPath, OFN_FILEMUSTEXIST, _T("Lynx files (*.evt)|Lynx.evt|"), this);
	if (IDOK == dlg.DoModal()) {
		CString szError;
		if (!ImportLinx(dlg.GetPathName(), szError))
			MessageBox(szError, _T("Import Competition Info"), MB_ICONERROR);
	}
}

void CSLLinxWatcherDlg::OnBnClickedDb2()
{
	CFileDialog dlg(TRUE, _T(".csv"), m_szPlayersPath, OFN_FILEMUSTEXIST, _T("CSV files (*.csv)|*.csv|"), this);
	if (IDOK == dlg.DoModal()) {
		CString szError;
		if (!ImportPlayers(dlg.GetPathName(), szError))
			MessageBox(szError, _T("Import Player's Info"), MB_ICONERROR);
	}
}

void CSLLinxWatcherDlg::OnBnClickedWatchPath()
{
	CFileDialog dlg(TRUE, _T(".csv"), m_szSrcResultCsvPath, OFN_FILEMUSTEXIST, _T("CSV-file with Race results (*.csv)|*.csv|"), this);
	if (IDOK == dlg.DoModal())
		m_szSrcResultCsvPath = dlg.GetPathName();
}

void CSLLinxWatcherDlg::OnBnClickedStartCsvPath()
{
	CFileDialog dlg(FALSE, _T(".csv"), m_szDstStartCsvPath, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST, _T("Start info CSV-file (*.csv)|*.csv|"), this);
	if (IDOK == dlg.DoModal())
		m_szDstStartCsvPath = dlg.GetPathName();
}

void CSLLinxWatcherDlg::OnBnClickedFinishCsvPath()
{
	CFileDialog dlg(FALSE, _T(".csv"), m_szDstFinishCsvPath, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST, _T("Finish info CSV-file (*.csv)|*.csv|"), this);
	if (IDOK == dlg.DoModal())
		m_szDstFinishCsvPath = dlg.GetPathName();
}

void CSLLinxWatcherDlg::OnBnClickedSaveAll()
{
	CFileDialog dlg(FALSE, _T(".SLLynxWatch"), m_szProjectPath, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST, _T(" Project file (*.SLLynxWatch)|*.SLLynxWatch|"), this);
	if (IDOK == dlg.DoModal()) {
		CString szError;
		if (!SaveProject(dlg.GetPathName(), szError))
			MessageBox(szError, _T("Save Project"), MB_ICONERROR);
	}
}

void CSLLinxWatcherDlg::OnBnClickedLoadAll()
{
	CFileDialog dlg(TRUE, _T(".SLLynxWatch"), m_szProjectPath, OFN_FILEMUSTEXIST, _T(" Project file (*.SLLynxWatch)|*.SLLynxWatch|"), this);
	if (IDOK == dlg.DoModal()) {
		CString szError;
		if (!LoadProject(dlg.GetPathName(), szError))
			MessageBox(szError, _T("Load Project"), MB_ICONERROR);
	}
}

void CSLLinxWatcherDlg::OnCbnSelchangeRace()
{
	FillPlayerList(GetCurRaceID());
}

void CSLLinxWatcherDlg::OnBnClickedSaveRace()
{
	CString szError;
	if (SaveStartCSV(m_szDstStartCsvPath, GetCurRaceID(), szError)) {
		SendTitlerCommand(0, MAKELONG(0, VK_F10));
	} else  MessageBox(szError, _T("Save Start List"), MB_ICONERROR);
}

void CSLLinxWatcherDlg::OnBnClickedSaveFinish()
{
	CString szError;
	if (SaveFinishCSV(m_szDstFinishCsvPath, GetCurRaceID(), szError)) {
		SendTitlerCommand(0, MAKELONG(0, VK_F11));
	} else  MessageBox(szError, _T("Save Finish List"), MB_ICONERROR);
	if (m_WatchResult.GetCheck() == BST_CHECKED) {
		m_WatchResult.SetCheck(BST_UNCHECKED);
		KillTimer(101);
		MessageBox(_T("READY !!!"), _T("Save Finish List"), MB_ICONEXCLAMATION);
	}
}

void CSLLinxWatcherDlg::OnBnClickedWatchResult()
{
	bool bRun = false;
	if (m_WatchResult.GetCheck() == BST_CHECKED && CheckResultCSV(m_szSrcResultCsvPath, false)) {
		if (HasResult(GetCurRaceID()))  OnBnClickedSaveFinish();
		else  bRun = true;
	}
	if (bRun)  SetTimer(101, 500, NULL);
	else {
		KillTimer(101);
		if (m_WatchResult.GetCheck() == BST_CHECKED)  m_WatchResult.SetCheck(BST_UNCHECKED);
	}
}

void CSLLinxWatcherDlg::OnTimer(UINT_PTR nIDEvent)
{
	if (CheckResultCSV(m_szSrcResultCsvPath, true) && HasResult(GetCurRaceID())) {
		OnBnClickedSaveFinish();
	}
	CDialogEx::OnTimer(nIDEvent);
}

void CSLLinxWatcherDlg::OnNMDblclkPlayerList(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMITEMACTIVATE pNMItemActivate = reinterpret_cast<LPNMITEMACTIVATE>(pNMHDR);
	int n = pNMItemActivate == NULL ? -1 : pNMItemActivate->iItem;
	if (n >= 0) {
		CString str = m_PlayerList.GetItemText(n, eBipColumn);
		int nBIP = _ttoi(str);
		for (sPlayerInfo & info : m_Players) if (info.m_nBIP == nBIP) {
			if (EditInfo(info)) {
				m_PlayerList.SetItemText(n, eFamilyColumn, info.m_szFamily);
				m_PlayerList.SetItemText(n, eNameColumn, info.m_szName);
			}
			break;
		}
	}
	*pResult = 0;
}

void CSLLinxWatcherDlg::OnBnClickedExit()
{
	if (MessageBox(_T("Close the aplication?"), szCaption, MB_ICONQUESTION | MB_YESNO) == IDYES)
		EndDialog(IDOK);
}

// -----------------
inline bool FindCoutnryName(int &id, char* &name)
{
	id = 0;
	while (*name >= '0' && *name <= '9') {
		id *= 10;
		id += *name++ - '0';
	}
	while (*name == ';' || *name == ',' || *name == '=' || *name == ' ' || *name == '-' || *name == ':')  name++;
	char *ptr = name;
	for (int i = 0; i < 3; i++)
		if (*ptr > 'A' && *ptr < 'Z')  ptr++;
		else if (*ptr > 'a' && *ptr < 'z')  *ptr++ -= 'a' - 'A';
		else  break;
	*ptr = 0;
	return id != 0 && *name != 0;
}

bool CSLLinxWatcherDlg::LoadCountryMap()
{
	FILE *file = nullptr;   char fname[MAX_PATH];
	GetModuleFileNameA(NULL, fname, MAX_PATH);
	CPathA path(fname);
	int nSkip = path.FindFileName();
	strcpy_s(&fname[nSkip], MAX_PATH - nSkip, "country.csv");
	if (fopen_s(&file, fname, "rt") != 0) {
		return false;
	}
	while (fgets(fname, MAX_PATH, file)) {
		int id=0;  char *name=fname;
		if(FindCoutnryName(id, name))
			m_nCountryMap.insert(std::make_pair(id, name));
	}
	fclose(file);
	return m_nCountryMap.size() > 0;
}

bool CSLLinxWatcherDlg::SaveProject(CString szPath, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"wt,ccs=UNICODE") != 0) {
		szError.Format(_T("Can't create file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	m_szProjectPath = szPath;
	// save Competition name
	CString szCompetition;
	m_Competion.GetWindowText(szCompetition);
	fputws(szCompetition + L"\r\n", file);
	// save paths
	fputws(m_szSrcLynxEvtPath + L"\r\n", file);
	fputws(m_szPlayersPath + L"\r\n", file);
	fputws(m_szSrcResultCsvPath + L"\r\n", file);
	fputws(m_szDstStartCsvPath + L"\r\n", file);
	fputws(m_szDstFinishCsvPath + L"\r\n", file);
	// save players
	for (const sPlayerInfo &info : m_Players) {
		CStringW str;
		// RaceID, Track, BIP, Country, Family, Name, Rank, ResultMS, DelayMS, ResultComment
		str.Format(L"%d,%d,%d,%s,%s,%s,%d,%d,%d,%s\n",
			info.m_nRaceID, info.m_nTrackIndex, info.m_nBIP, info.m_szCountry,
			info.m_szFamily, info.m_szName, info.m_nRank, info.m_nResultMS,
			info.m_nDelayMS, info.m_szResultComment);
		fputws(str, file);
	}
	fclose(file);
	return true;
}

inline bool ScanEvtData(wchar_t* &ptr)
{
	while (*ptr != 0) {
		if (*ptr == L',' || *ptr == L'\n' || *ptr == L'\r') {
			*ptr++ = 0;
			return true;
		}
		ptr++;
	}
	return false;
}

inline bool ScanProjectLine(FILE *file, wchar_t tmp[4096])
{
	wchar_t *ptr = tmp;
	return fgetws(tmp, 4096, file) != nullptr && ScanEvtData(ptr);
}

bool CSLLinxWatcherDlg::LoadProject(CString szPath, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"rt,ccs=UNICODE") != 0) {
		szError.Format(_T("Can't open file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	m_szProjectPath = szPath;
	wchar_t tmp[4096];
	// read Competition name
	if (ScanProjectLine(file, tmp))  m_Competion.SetWindowText(tmp);
	// read paths
	if (ScanProjectLine(file, tmp))  m_szSrcLynxEvtPath = tmp;
	if (ScanProjectLine(file, tmp))  m_szPlayersPath = tmp;
	if (ScanProjectLine(file, tmp))  m_szSrcResultCsvPath = tmp;
	if (ScanProjectLine(file, tmp))  m_szDstStartCsvPath = tmp;
	if (ScanProjectLine(file, tmp))  m_szDstFinishCsvPath = tmp;
	// read players
	m_Players.clear();
	while (fgetws(tmp, 4096, file) != nullptr) {
		wchar_t *ptr = tmp;
		// RaceID
		wchar_t *szRaceId = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// Track
		wchar_t *szTrack = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// BIP
		wchar_t *szBIP = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// CountryID
		wchar_t *szCountryID = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// Family
		wchar_t *szFamily = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// Name
		wchar_t *szName = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// RankID
		wchar_t *szRankID = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// ResultMS
		wchar_t *szResultMS = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// DelayMS
		wchar_t *szDelayMS = ptr;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		// ResultComment
		wchar_t *szResultComment = ptr;
		ScanEvtData(ptr);
		// add player
		sPlayerInfo info;
		info.m_nRaceID = _wtoi(szRaceId);
		info.m_nTrackIndex = _wtoi(szTrack);
		info.m_nBIP = _wtoi(szBIP);
		_tcscpy_s(info.m_szCountry, szCountryID);
		info.m_szFamily = szFamily;
		info.m_szName = szName;
		info.m_nRank = _wtoi(szRankID);
		info.m_nResultMS = _wtoi(szResultMS);
		info.m_nDelayMS = _wtoi(szDelayMS);
		info.m_szResultComment = szResultComment;
		m_Players.push_back(info);
	}
	fclose(file);
	// update info
	m_RaceList.ResetContent();
	int nLast = -1;
	TCHAR str[10];
	for (auto info : m_Players) if (info.m_nRaceID != nLast) {
		nLast = info.m_nRaceID;
		_itot_s(nLast, str, 10);
		m_RaceList.AddString(str);
	}
	m_PlayerList.DeleteAllItems();
	m_RaceList.SetCurSel(0);
	OnCbnSelchangeRace();
	return true;
}
//
inline bool SplitName(wchar_t* &ptr)
{
	while (*ptr != 0) {
		if (*ptr == L' ' || *ptr == L'_') {
			*ptr++ = 0;
			return true;
		}
		ptr++;
	}
	return false;
}
//
bool CSLLinxWatcherDlg::ImportLinx(CString szPath, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"rt,ccs=UTF-16LE") != 0) {
		szError.Format(_T("Can't open file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	// file exist -> reset local info & fill it from the file
	m_Players.clear();
	m_RaceList.ResetContent();
	m_PlayerList.DeleteAllItems();
	m_szSrcLynxEvtPath = szPath;
	// read the text
	wchar_t tmp[4096];  int nRaceIndex = 0;
	while (fgetws(tmp, 4096, file) != nullptr) {
		wchar_t *ptr = tmp;
		if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
		if (tmp[0] != 0) { // race header
			// skip second column
			if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
			// 3-d column = race index
			wchar_t *szRaceId = ptr;
			if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
			// 4-d column = competion name
			wchar_t *szCompetitionName = ptr;
			ScanEvtData(ptr);
			// all right - use the race id
			nRaceIndex = _wtoi(szRaceId);
			m_Competion.SetWindowText(szCompetitionName);
			m_RaceList.AddString(szRaceId);
		} else { // sportsmen info
			// 2-d column = track number
			wchar_t *szTrackId = ptr;
			if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
			// skip 3-d & 4-th columns
			if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
			if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
			// 5-th column = family and name
			wchar_t *szName = ptr;
			if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
			// 6-th column = BIP
			wchar_t *szBipId = ptr;
			if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
			// skip 7-th column
			if (!ScanEvtData(ptr)) { ASSERT(false);  continue; }
			// 8-d column = country id
			wchar_t *szCountryId = ptr;
			ScanEvtData(ptr);
			// all right - use the race id
			sPlayerInfo info;
			info.m_nRaceID = nRaceIndex;
			info.m_nTrackIndex = _wtoi(szTrackId);
			info.m_nBIP = _wtoi(szBipId);
			GetCountry(_wtoi(szCountryId), info.m_szCountry);
			ptr = szName;
			SplitName(ptr);
			info.m_szFamily = szName;
			info.m_szName = ptr;
			info.m_nRank = 0;
			info.m_nResultMS = 0;
			info.m_nDelayMS = 0;
			m_Players.push_back(info);
		}
	}
	fclose(file);
	// fill player's info
	m_RaceList.SetCurSel(0);
	OnCbnSelchangeRace();
	return true;
}

inline bool ScanCsvData(char* &ptr, char* &data)
{
	char nEndChar = 0;
	if (*ptr == '\'' || *ptr == '\"')  nEndChar = *ptr++;
	data = ptr;
	while (*ptr != 0) {
		if (*ptr == '\n' || *ptr == '\r') {
			*ptr++ = 0;
			return true;
		} else if (*ptr == nEndChar) {
			nEndChar = 0;
			*ptr = 0;
		} else if (*ptr == ',' || *ptr == ';' || *ptr == '\t') {
			*ptr++ = 0;
			return true;
		}
		ptr++;
	}
	return false;
}

bool CSLLinxWatcherDlg::ImportPlayers(CString szPath, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"rt") != 0) {
		szError.Format(_T("Can't open file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	// file exist -> reset local info & fill it from the file
	m_szPlayersPath = szPath;
	// skip caption
	char tmp[4096];
	fgets(tmp, 4096, file);
	// read the text
	while (fgets(tmp, 4096, file) != nullptr) {
		char *ptr = tmp;
		char *szBip, *szCountry, *szFullName, *szFlag;
		if (!ScanCsvData(ptr, szBip)) { ASSERT(false);  continue; }
		if (!ScanCsvData(ptr, szCountry)) { ASSERT(false);  continue; }
		if (!ScanCsvData(ptr, szFullName)) { ASSERT(false);  continue; }
		if (!ScanCsvData(ptr, szFlag)) { ASSERT(false);  continue; }
		// all right - change player data
		int nBIP = atoi(szBip);
		for (sPlayerInfo & info : m_Players) if (info.m_nBIP == nBIP) {
			for(int i=0; i<3; i++)  info.m_szCountry[i] = szCountry[i];
			info.m_szCountry[4] = 0;
			char *rest = szFullName;
			while (*rest != 0) if (*rest++ == ' ') {
				*(rest - 1) = 0;
				break;
			}
			info.m_szFamily = szFullName;
			info.m_szName = rest;
		}
	}
	fclose(file);
	// fill player's info
	m_RaceList.SetCurSel(0);
	OnCbnSelchangeRace();
	return true;
}

void CSLLinxWatcherDlg::GetCountry(int id, TCHAR szCountry[5])
{
	szCountry[0] = 0;
	auto iter = m_nCountryMap.find(id);
	if (iter == m_nCountryMap.cend())  return;
	char name[4] = { 0 };
	strncpy_s(name, (*iter).second.c_str(), 3);
	for (int i = 0; i < 3; i++)  szCountry[i] = name[i];
	szCountry[3] = 0;
}

int CSLLinxWatcherDlg::GetCurRaceID()
{
	CString str;
	int n = m_RaceList.GetCurSel();
	m_RaceList.GetLBText(n, str);
	return _ttoi(str);
}

void CSLLinxWatcherDlg::FillPlayerList(int nRaceID)
{
	TCHAR szTrack[10], szBIP[10], szRank[20], szTime[20];
	m_PlayerList.DeleteAllItems();
	for (auto info : m_Players) if(info.m_nRaceID == nRaceID) {
		_itot_s(info.m_nTrackIndex, szTrack, 10);
		_itot_s(info.m_nBIP, szBIP, 10);
		_itot_s(info.m_nResultMS, szTime, 10);
		if (info.m_nRank == 0)  _tcscpy_s(szRank, info.m_szResultComment);
		else  _itot_s(info.m_nRank, szRank, 10);
		int n = m_PlayerList.InsertItem(m_PlayerList.GetItemCount(), _T(""));
		m_PlayerList.SetItemText(n, eTrackColumn, szTrack);
		m_PlayerList.SetItemText(n, eRankColumn, szRank);
		m_PlayerList.SetItemText(n, eBipColumn, szBIP);
		m_PlayerList.SetItemText(n, eFamilyColumn, info.m_szFamily);
		m_PlayerList.SetItemText(n, eNameColumn, info.m_szName);
		m_PlayerList.SetItemText(n, eTimeColumn, szTime);
		m_PlayerList.SetItemText(n, eCountryColumn, info.m_szCountry);
	}
}

bool CSLLinxWatcherDlg::HasResult(int nRaceID)
{
	for (auto info : m_Players) if (info.m_nRaceID == nRaceID) {
		if (info.m_nRank == 0 && info.m_szResultComment.IsEmpty())
			return false;
	}
	return true;
}

inline bool ScanResultData(char* &ptr)
{
	while (*ptr != 0) {
		if (*ptr == ',' || *ptr == '\n' || *ptr == '\r') {
			if (*(ptr - 1) == '\"' || *(ptr - 1) == '\'')  *(ptr - 1) = 0;
			*ptr++ = 0;
			return true;
		}
		ptr++;
	}
	return false;
}

bool CSLLinxWatcherDlg::CheckResultCSV(CString szPath, bool bCheckTime)
{
	WIN32_FILE_ATTRIBUTE_DATA data;
	GET_FILEEX_INFO_LEVELS fInfoLevelId = GetFileExInfoStandard;
	if (GetFileAttributesEx(szPath, fInfoLevelId, &data) == 0)
		return false;
	if (bCheckTime) {
		if (data.ftLastWriteTime.dwHighDateTime == m_ftResultCsv.dwHighDateTime &&
			data.ftLastWriteTime.dwLowDateTime == m_ftResultCsv.dwLowDateTime)
			return true;
	}
	// read file
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"rt") != 0)  return false;
	char tmp[4096];
	while (fgets(tmp, 4096, file) != nullptr) {
		char *ptr = tmp;
		// skip 5 columns
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		// 6: BIP
		char *szBip = ++ptr;
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		// skip 14-7=7 columns
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		// 14: time
		char *szTime = ++ptr;
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		// 15: delay
		char *szDelay = ++ptr;
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		// 16: rank
		char *szRank = ++ptr;
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		// 17: comment
		char *szComment = ++ptr;
		if (!ScanResultData(ptr)) { ASSERT(false);  continue; }
		// correct data
		CString str(szComment);
		SetResultInfo(atoi(szBip), atoi(szRank), atoi(szTime), atoi(szDelay), str);
	}
	fclose(file);
	// save last modification time
	m_ftResultCsv.dwHighDateTime = data.ftLastWriteTime.dwHighDateTime;
	m_ftResultCsv.dwLowDateTime = data.ftLastWriteTime.dwLowDateTime;
	// redraw info
	OnCbnSelchangeRace();
	return true;
}

bool CSLLinxWatcherDlg::SetResultInfo(int nBip, int nRank, int nTimeMS, int nDelayMS, CString szComment)
{
	for (sPlayerInfo & info : m_Players) if (info.m_nBIP == nBip) {
		info.m_nRank = nRank;
		info.m_nResultMS = nTimeMS;
		info.m_nDelayMS = nDelayMS;
		info.m_szResultComment = szComment;
		return true;
	}
	return false;
}

bool CompByTrack(const sPlayerInfo &a, const sPlayerInfo &b)
{
	return a.m_nTrackIndex < b.m_nTrackIndex;
}

bool CompByRank(const sPlayerInfo &a, const sPlayerInfo &b)
{
	if (a.m_nRank > 0 && b.m_nRank > 0)  return a.m_nRank < b.m_nRank;
	else if (a.m_nRank == 0 && b.m_nRank == 0)  return a.m_nTrackIndex < b.m_nTrackIndex;
	else  return a.m_nRank > 0;
}

bool CSLLinxWatcherDlg::SaveStartCSV(CString szPath, int nRaceID, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"wt,ccs=UNICODE") != 0) {
		szError.Format(_T("Can't create file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	CString szCompetition;
	m_Competion.GetWindowText(szCompetition);
	// save header
	fputws(L"Race,Track,BIP,Name,Country,Flag\n", file);
	// select & sort players
	std::vector<sPlayerInfo> race;
	std::copy_if(m_Players.begin(), m_Players.end(), back_inserter(race), [nRaceID](const sPlayerInfo &item) { return item.m_nRaceID == nRaceID; });
	std::sort(race.begin(), race.end(), CompByTrack);
	// save players
	for (const sPlayerInfo &info : race) {
		CStringW str;
		str.Format(L"%s,%d,%d,%s %s,%s,Base\\Flags\\%s.png\n",
			szCompetition, info.m_nTrackIndex, info.m_nBIP, info.m_szFamily, info.m_szName, info.m_szCountry, info.m_szCountry);
		fputws(str, file);
	}
	fclose(file);
	return true;
}

bool CSLLinxWatcherDlg::SaveFinishCSV(CString szPath, int nRaceID, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"wt,ccs=UNICODE") != 0) {
		szError.Format(_T("Can't create file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	CString szCompetition;
	m_Competion.GetWindowText(szCompetition);
	// save header
	fputws(L"Race,Rank,BIP,Name,Country,Flag,Comment,Time,Diff\n", file);
	// select & sort players
	std::vector<sPlayerInfo> race;
	std::copy_if(m_Players.begin(), m_Players.end(), back_inserter(race), [nRaceID](const sPlayerInfo &item) { return item.m_nRaceID == nRaceID; });
	std::sort(race.begin(), race.end(), CompByRank);
	// save players
	for (const sPlayerInfo &info : race) {
		CStringW str, szTime, szDiff;  WCHAR szRank[20];
		if (info.m_nRank == 0)  szRank[0] = 0;
		else  _itow_s(info.m_nRank, szRank, 10);
		int nSec = info.m_nResultMS / 1000;
		int nMS = info.m_nResultMS - nSec * 1000;
		int nMin = nSec / 60;
		nSec -= nMin * 60;
		szTime.Format(L"%d:%02d.%03d", nMin, nSec, nMS);
		if (info.m_nRank != 1) {
			if (info.m_nDelayMS == 0)  szDiff = _T("");
			else if (info.m_nDelayMS < 1000)  szDiff.Format(L"+0.%03d", info.m_nDelayMS);
			else if(info.m_nDelayMS < 60*1000)  szDiff.Format(L"+%d.%03d", info.m_nDelayMS / 1000, info.m_nDelayMS % 1000);
			else {
				nSec = info.m_nDelayMS / 1000;
				nMS = info.m_nDelayMS - nSec * 1000;
				nMin = nSec / 60;
				nSec -= nMin * 60;
				szDiff.Format(L"+%d:%02d.%03d", nMin, nSec, nMS);
			}
		} else  szDiff = szTime;
		str.Format(L"%s,%s,%d,%s %s,%s,Base\\Flags\\%s.png,%s,%s,%s\n",
			szCompetition, szRank, info.m_nBIP, info.m_szFamily, info.m_szName, info.m_szCountry, info.m_szCountry, info.m_szResultComment, szTime, szDiff);
		fputws(str, file);
	}
	fclose(file);
	return true;
}

bool CSLLinxWatcherDlg::EditInfo(sPlayerInfo & info)
{
	TCHAR szBip[10], szTime[20];
	_itot_s(info.m_nBIP, szBip, 10);
	_itot_s(info.m_nResultMS, szTime, 10);
	CEditInfoDlg dlg(this, szBip, info.m_szFamily, info.m_szName, szTime, info.m_szResultComment);
	if (dlg.DoModal() == IDOK) {
		info.m_szFamily = dlg.Family();
		info.m_szName = dlg.Name();
		if (dlg.RankChanged()) {
			info.m_szResultComment = dlg.Comment();
			info.m_nResultMS = _ttoi(dlg.Time());
			DWORD nBestTime;
			RecalcRank(info.m_nRaceID, nBestTime);
			info.m_nDelayMS = info.m_nResultMS - nBestTime;
			OnCbnSelchangeRace();
		}
		return true;
	}
	return false;
}

bool CompByTime(const sPlayerInfo &a, const sPlayerInfo &b)
{
	if (a.m_nResultMS > 0 && b.m_nResultMS > 0)  return a.m_nResultMS < b.m_nResultMS;
	else if (a.m_nResultMS == 0 && b.m_nResultMS == 0)  return a.m_nTrackIndex < b.m_nTrackIndex;
	else  return a.m_nResultMS > 0;
}

void CSLLinxWatcherDlg::RecalcRank(int nRaceID, DWORD &nBestTime)
{
	// select the race and sort it
	std::vector<sPlayerInfo> race;
	std::copy_if(m_Players.begin(), m_Players.end(), back_inserter(race), [nRaceID](const sPlayerInfo &item) { return item.m_nRaceID == nRaceID; });
	std::sort(race.begin(), race.end(), CompByTime);
	// fill rank
	int nRank = 0;
	nBestTime = 0;
	for (sPlayerInfo &info : race) {
		if (info.m_szResultComment.IsEmpty())  info.m_nRank = ++nRank;
		else  info.m_nRank = 0;
		if (info.m_nRank == 1)  nBestTime = info.m_nResultMS;
		if (info.m_nResultMS == 0) info.m_nDelayMS = 0;
		else  info.m_nDelayMS = info.m_nResultMS - nBestTime;
		for (sPlayerInfo &src : m_Players) if (src.m_nRaceID == nRaceID && src.m_nBIP == info.m_nBIP) {
			src.m_nRank = info.m_nRank;
			src.m_nDelayMS = info.m_nDelayMS;
			break;
		}
	}
}

struct CommandInfo {
	WPARAM wParam;
	LPARAM lParam;
};

BOOL CALLBACK SendTitleMessage(HWND hwnd, LPARAM lParam)
{
	static const TCHAR szCaption[] = _T("SLSportTitler");
	static const int lnCaption = _countof(szCaption) - 1;
	TCHAR name[MAX_PATH];
	int ln = GetWindowText(hwnd, name, MAX_PATH);
	if (ln >= lnCaption && _tcsncicmp(name, szCaption, lnCaption)==0) {
		CommandInfo *info = (CommandInfo*)lParam;
		::PostMessage(hwnd, WM_HOTKEY, info->wParam, info->lParam);
		return FALSE;
	}
	return TRUE;
}

void CSLLinxWatcherDlg::SendTitlerCommand(WPARAM wParam, LPARAM lParam)
{
	CommandInfo info;
	info.wParam = wParam;
	info.lParam = lParam;
	EnumWindows(SendTitleMessage, (LPARAM)(void*)&info);
}
