
// SLTramplinWatcherDlg.cpp : implementation file
//

#include "stdafx.h"
#include "SLTramplinWatcher.h"
#include "SLTramplinWatcherDlg.h"
#include "TCPSettingsDlg.h"
#include "PlayerInfo.h"
#include "afxdialogex.h"
#include <algorithm>
#include <iterator>
#include <atlpath.h>
#include <utility>
#include <locale>
#include <codecvt>

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

const int WM_TCP_DATA = WM_USER + 1;

const int cTitlerCmd4SpeedOn = VK_F10;
const int cTitlerCmd4SpeedOff = VK_F10;
const int cTitlerCmd4DistanceOn = VK_F11;
const int cTitlerCmd4DistanceOff = VK_F11;
const int cTimeout4SpeedOff = 3000;
const int cTimeout4DistanceOff = 3000;
const int cTimeout4LastPlayer = 500;
const int cTimer4SpeedOff = 201;
const int cTimer4DistanceOff = 202;
const int cTimer4LastPlayer = 203;


// CSLTramplinWatcherDlg dialog

CSLTramplinWatcherDlg::CSLTramplinWatcherDlg(CWnd* pParent /*=NULL*/)
	: CDialogEx(IDD_SLTRAMPLINWATCHER_DIALOG, pParent)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
	m_nIP = 0x7F000001;
	m_nPort = 25000;
	m_sSocket = INVALID_SOCKET;
	m_pCurrentInfo = nullptr;
	m_pLastInfo = nullptr;
	m_bUpdateConnectionInfo = false;
	m_bUpdateCurrentInfo = false;
	m_bCheckCurrentInfoFullness = false;
	m_bRedrawPlayersInfo = false;
}
// save/restore global settings
static const TCHAR *szCaption = _T("SLTramplinWatcher");
static const TCHAR *szSection = _T("");
static const TCHAR *szProjectPath = _T("Project_Path");
static const TCHAR *szPlayersPath = _T("PlayersCSV_Path");
static const TCHAR *szResultPath = _T("ResultCSV_Path");
static const TCHAR *szIP = _T("IP");
static const TCHAR *szPort = _T("Port");
//
void CSLTramplinWatcherDlg::SaveSettings(CWinApp *pApp)
{
	pApp->WriteProfileString(szSection, szProjectPath, m_szProjectPath);
	pApp->WriteProfileString(szSection, szPlayersPath, m_szPlayersPath);
	pApp->WriteProfileString(szSection, szResultPath, m_szResultPath);
	pApp->WriteProfileInt(szSection, szIP, m_nIP);
	pApp->WriteProfileInt(szSection, szPort, m_nPort);
}
//
void CSLTramplinWatcherDlg::LoadSettings(CWinApp *pApp)
{
	m_szProjectPath = pApp->GetProfileString(szSection, szProjectPath, m_szProjectPath);
	m_szPlayersPath = pApp->GetProfileString(szSection, szPlayersPath, m_szPlayersPath);
	m_szResultPath = pApp->GetProfileString(szSection, szResultPath, m_szResultPath);
	m_nIP = pApp->GetProfileInt(szSection, szIP, m_nIP);
	m_nPort = pApp->GetProfileInt(szSection, szPort, m_nPort);
}

void CSLTramplinWatcherDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialogEx::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_PLAYERS, m_PlayerList);
	DDX_Control(pDX, IDC_RACE, m_RaceList);
	DDX_Control(pDX, IDC_TCP_WATCH, m_WatchTCP);
	DDX_Control(pDX, IDC_RUN2, m_Run2);
	DDX_Control(pDX, IDC_TABLE_CNT, m_TableCnt);
	DDX_Control(pDX, IDC_TCP_STATE, m_TCPInfo);
	DDX_Control(pDX, IDC_CURRENT_INFO, m_CurrentInfo);
	DDX_Control(pDX, IDC_MARK_CURRENT, m_MarkCurrent);
}

BEGIN_MESSAGE_MAP(CSLTramplinWatcherDlg, CDialogEx)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_EXIT, &CSLTramplinWatcherDlg::OnBnClickedExit)
	ON_BN_CLICKED(IDC_IMPORT, &CSLTramplinWatcherDlg::OnBnClickedImport)
	ON_BN_CLICKED(IDC_SAVE_PROJECT, &CSLTramplinWatcherDlg::OnBnClickedSaveProject)
	ON_BN_CLICKED(IDC_LOAD_PROJECT, &CSLTramplinWatcherDlg::OnBnClickedLoadProject)
	ON_BN_CLICKED(IDC_TCP_SETTINGS, &CSLTramplinWatcherDlg::OnBnClickedTcpSettings)
	ON_BN_CLICKED(IDC_TCP_WATCH, &CSLTramplinWatcherDlg::OnBnClickedTcpWatch)
	ON_CBN_SELCHANGE(IDC_RACE, &CSLTramplinWatcherDlg::OnCbnSelchangeRace)
	ON_BN_CLICKED(IDC_RUN2, &CSLTramplinWatcherDlg::OnBnClickedRun2)
	ON_BN_CLICKED(IDC_RESULT_PATH, &CSLTramplinWatcherDlg::OnBnClickedResultPath)
	ON_NOTIFY(NM_DBLCLK, IDC_PLAYERS, &CSLTramplinWatcherDlg::OnNMDblclkPlayers)
	ON_BN_CLICKED(IDC_MAKE_RESULT, &CSLTramplinWatcherDlg::OnBnClickedMakeResult)
	ON_BN_CLICKED(IDC_MAKE_START, &CSLTramplinWatcherDlg::OnBnClickedMakeStart)
	ON_BN_CLICKED(IDC_MAKE_CURRENT, &CSLTramplinWatcherDlg::OnBnClickedMakeLast)
	ON_MESSAGE(WM_TCP_DATA, &CSLTramplinWatcherDlg::OnTCPData)
END_MESSAGE_MAP()


// CSLTramplinWatcherDlg message handlers

BOOL CSLTramplinWatcherDlg::OnInitDialog()
{
	CDialogEx::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		BOOL bNameValid;
		CString strAboutMenu;
		bNameValid = strAboutMenu.LoadString(IDS_ABOUTBOX);
		ASSERT(bNameValid);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	if (!LoadRaces()) {
		MessageBox(_T("There is no file with race list!\r\n\r\nPut file \'race.csv\' (UNICODE) near EXE-file.\r\n\r\nFile must have 3 colums: RaceIndex, Woman(1/0), RaceName"), szCaption, MB_ICONSTOP);
		EndDialog(IDCANCEL);
	}

	// Init locals
	RECT rc;  int cx[eLastColumn];
	m_PlayerList.GetClientRect(&rc);
	rc.right -= rc.left;
	cx[ePosColumn] = rc.right / 25;
	cx[eRankColumn] = rc.right / 12;
	cx[eNumberColumn] = rc.right / 13;
	cx[eRegionColumn] = rc.right / 9;
	cx[eScoreColumn] = rc.right / 9;
	rc.right -= cx[ePosColumn] + cx[eRankColumn] + cx[eNumberColumn] + cx[eScoreColumn] + cx[eRegionColumn];
	cx[eNameColumn] = rc.right;
	m_PlayerList.SetExtendedStyle(LVS_EX_GRIDLINES | LVS_EX_FULLROWSELECT);
	m_PlayerList.InsertColumn(ePosColumn, _T("#"), LVCFMT_CENTER, cx[ePosColumn]);
	m_PlayerList.InsertColumn(eRankColumn, _T("Rank"), LVCFMT_CENTER, cx[eRankColumn]);
	m_PlayerList.InsertColumn(eNumberColumn, _T("BIP"), LVCFMT_CENTER, cx[eNumberColumn]);
	m_PlayerList.InsertColumn(eNameColumn, _T("Name"), LVCFMT_LEFT, cx[eNameColumn]);
	m_PlayerList.InsertColumn(eScoreColumn, _T("Score"), LVCFMT_LEFT, cx[eScoreColumn]);
	m_PlayerList.InsertColumn(eRegionColumn, _T("Country"), LVCFMT_CENTER, cx[eRegionColumn]);

	m_RaceList.SetCurSel(0);
	m_Run2.SetCheck(BST_UNCHECKED);

	SetTimer(101, 100, NULL);

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CSLTramplinWatcherDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CDialogEx dlgAbout(IDD_ABOUTBOX);
		dlgAbout.DoModal();
	}
	else
	{
		CDialogEx::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CSLTramplinWatcherDlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialogEx::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CSLTramplinWatcherDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

void CSLTramplinWatcherDlg::OnBnClickedExit()
{
	EndDialog(IDOK);
}

void CSLTramplinWatcherDlg::OnBnClickedImport()
{
	CFileDialog dlg(TRUE, _T(".csv"), m_szPlayersPath, OFN_FILEMUSTEXIST, _T("CSV files (*.csv)|*.csv|"), this);
	if (IDOK == dlg.DoModal()) {
		CString szError;
		if (!ImportPlayers(dlg.GetPathName(), szError))
			MessageBox(szError, _T("Import Player's Info"), MB_ICONERROR);
	}
}

void CSLTramplinWatcherDlg::OnBnClickedSaveProject()
{
	CFileDialog dlg(FALSE, _T(".SLTrampWatch"), m_szProjectPath, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST, _T(" Project file (*.SLTrampWatch)|*.SLTrampWatch|"), this);
	if (IDOK == dlg.DoModal()) {
		CString szError;
		if (!SaveProject(dlg.GetPathName(), szError))
			MessageBox(szError, _T("Save Project"), MB_ICONERROR);
	}
}

void CSLTramplinWatcherDlg::OnBnClickedLoadProject()
{
	CFileDialog dlg(TRUE, _T(".SLTrampWatch"), m_szProjectPath, OFN_FILEMUSTEXIST, _T(" Project file (*.SLTrampWatch)|*.SLTrampWatch|"), this);
	if (IDOK == dlg.DoModal()) {
		CString szError;
		if (!LoadProject(dlg.GetPathName(), szError))
			MessageBox(szError, _T("Load Project"), MB_ICONERROR);
	}
}

void CSLTramplinWatcherDlg::OnBnClickedTcpSettings()
{
	CTCPSettingsDlg dlg(this, m_nIP, m_nPort);
	if (dlg.DoModal() == IDOK) {
		m_nIP = dlg.IP();
		m_nPort = dlg.Port();
		OnBnClickedTcpWatch();
	}
}

void CSLTramplinWatcherDlg::OnBnClickedTcpWatch()
{
	if (m_WatchTCP.GetCheck() == BST_CHECKED)  OnStartTCPWatch();
	else  OnStopTCPWatch();
}

void CSLTramplinWatcherDlg::OnCbnSelchangeRace()
{
	FillPlayerList();
}

void CSLTramplinWatcherDlg::OnBnClickedResultPath()
{
	CFileDialog dlg(FALSE, _T(".csv"), m_szResultPath, OFN_HIDEREADONLY | OFN_PATHMUSTEXIST, _T("Result files (*.csv)|*.csv|"), this);
	if (IDOK == dlg.DoModal()) {
		m_szResultPath = dlg.GetPathName();
	}
}

void CSLTramplinWatcherDlg::OnBnClickedMakeStart()
{
	CString szError;
	if (!SaveStartTables(szError))
		MessageBox(szError, _T("Save Start Tables"), MB_ICONERROR);
}

void CSLTramplinWatcherDlg::OnBnClickedMakeResult()
{
	CString szError;
	if (!SaveResultTables(szError))
		MessageBox(szError, _T("Save Result Tables"), MB_ICONERROR);
}

void CSLTramplinWatcherDlg::OnBnClickedMakeLast()
{
	if (nullptr == m_pLastInfo)  return;
	CString szError;
	if (!SavePlayer(m_pLastInfo, true, szError))
		MessageBox(szError, _T("Save Last Player Info"), MB_ICONERROR);
}

void CSLTramplinWatcherDlg::OnBnClickedRun2()
{
	FillPlayerList();
	//OnBnClickedMakeResult();
}

void CSLTramplinWatcherDlg::OnNMDblclkPlayers(NMHDR *pNMHDR, LRESULT *pResult)
{
	LPNMITEMACTIVATE pNMItemActivate = reinterpret_cast<LPNMITEMACTIVATE>(pNMHDR);
	int n = pNMItemActivate == NULL ? -1 : pNMItemActivate->iItem;
	if (n >= 0) {
		CString str = m_PlayerList.GetItemText(n, eNumberColumn);
		int nNumber = _ttoi(str);
		sPlayerInfo *info = GetPlayer(nNumber);
		CPlayerInfo dlg(this, info);
		if (info != nullptr && dlg.DoModal() == IDOK) {
			FillPlayerList();
			if (dlg.SelectAsCurrent()) {
				m_pLastInfo = info;
				m_pCurrentInfo = info;
				m_bUpdateCurrentInfo = true;
			}
		}
	}
	*pResult = 0;
}

void CSLTramplinWatcherDlg::OnTimer(UINT_PTR nIDEvent)
{
	if (nIDEvent == cTimer4SpeedOff) {
		SendSLTitlerCmd(cTitlerCmd4SpeedOff);
		KillTimer(cTimer4SpeedOff);
	} else if (nIDEvent == cTimer4DistanceOff) {
		SendSLTitlerCmd(cTitlerCmd4DistanceOff);
		KillTimer(cTimer4DistanceOff);
	} else if (nIDEvent == cTimer4LastPlayer) {
		static bool bInTimer = false;
		m_MarkCurrent.ShowWindow(FALSE);
		KillTimer(cTimer4LastPlayer);
		CString szError;
		if (!SavePlayer(m_pLastInfo, true, szError) && !bInTimer) {
			bInTimer = true;
			MessageBox(szError, _T("Save Last Player Info"), MB_ICONERROR);
			bInTimer = false;
		}
	} else {
		static bool bInTimer = false;
		if (!bInTimer) {
			bInTimer = true;
			if (m_bUpdateConnectionInfo) {
				m_bUpdateConnectionInfo = false;
				CString str;
				if (m_bWachig)  str.Format(_T("TCP data size = %d"), m_nTotalDataSize);
				else  str = _T("TCP not connected");
				SetDlgItemText(IDC_TCP_STATE, str);
			}
			if (m_bUpdateCurrentInfo) {
				m_bUpdateCurrentInfo = false;
				CString str;
				if (nullptr != m_pCurrentInfo) {
					CString szError;
					if (!SavePlayer(m_pCurrentInfo, false, szError))
						MessageBox(szError, _T("Save Current Player Info"), MB_ICONERROR);
					str.Format(_T("BIP:%d  Speed=%s/%s  Distance=%s/%s  Score=%s/%s"),
						m_pCurrentInfo->m_nNumber, m_pCurrentInfo->m_szSpeed[0], m_pCurrentInfo->m_szSpeed[1],
						m_pCurrentInfo->m_szDistance[0], m_pCurrentInfo->m_szDistance[1],
						Score2Text(m_pCurrentInfo->m_nScore[0]), Score2Text(m_pCurrentInfo->m_nScore[1]));
				} else {
					str = _T("No current player");
				}
				SetDlgItemText(IDC_CURRENT_INFO, str);
			}
			if (m_bCheckCurrentInfoFullness) {
				m_bCheckCurrentInfoFullness = false;
				// EGOR - just wait rank from next 'SAVE'
				//if (CurrentIsReady()) {
				//	OnBnClickedMakeCurrent();
				//}
			}
			if (m_bRedrawPlayersInfo) {
				m_bRedrawPlayersInfo = false;
				FillPlayerList();
			}
			bInTimer = false;
		}
	}
	CDialogEx::OnTimer(nIDEvent);
}

LRESULT CSLTramplinWatcherDlg::OnTCPData(WPARAM size, LPARAM data)
{
	if (size > 0) { // check data
		m_bUpdateConnectionInfo = true;
		m_nTotalDataSize += size;
		ParseData((char*)data, size);
		return TRUE;
	}
	else if (size == 0) {
		MessageBox(_T("TCP server stopped !!!"), szCaption, MB_ICONEXCLAMATION);
		OnStopTCPWatch();
	}
	return FALSE;
}


// -----------------------
//
inline bool ScanData(wchar_t* &ptr)
{
	while (*ptr != 0) {
		if (*ptr == L',' || *ptr == L'\n' || *ptr == L'\r') {
			*ptr++ = 0;
			return true;
		}
		ptr++;
	}
	return false;
}
//
bool CSLTramplinWatcherDlg::LoadRaces()
{
	FILE *file = nullptr;  wchar_t fname[MAX_PATH];
	GetModuleFileName(NULL, fname, MAX_PATH);
	CPath path(fname);
	int nSkip = path.FindFileName();
	_tcscpy_s(&fname[nSkip], MAX_PATH - nSkip, L"race.csv");
	if (_wfopen_s(&file, fname, L"rt,ccs=UNICODE") != 0) {
		return false;
	}
	// read players
	wchar_t tmp[4096];
	while (fgetws(tmp, 4096, file) != nullptr) {
		wchar_t *ptr = tmp;
		// nRace
		wchar_t *szRace = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// bWoman
		wchar_t *szWoman = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// sName
		wchar_t *szName = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// add race
		sRaceInfo info;
		info.m_nRace = _wtoi(szRace);
		info.m_bWoman = _wtoi(szWoman) != 0;
		info.m_szName = szName;
		m_Races.push_back(info);
		// fill race list
		CString szRaceName;  szRaceName.Format(_T("%d:%s"), info.m_nRace, info.m_szName);
		m_RaceList.AddString(szRaceName);
	}
	fclose(file);
	return m_Races.size() > 0;
}
//
bool CSLTramplinWatcherDlg::SaveProject(CString szPath, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"wt,ccs=UNICODE") != 0) {
		szError.Format(_T("Can't create file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	m_szProjectPath = szPath;
	// save players
	for (const sPlayerInfo &info : m_Players) {
		CStringW str;
		// nNumber, sRegion, sFlag, sName, bWoman, nRace, nPos, sTCPName, sSpeed[cTryCnt], sDistance[cTryCnt], sWind[cTryCnt], nRank[cTryCnt], sScore[cTryCnt], sReferee[cTryCnt*cRefereeCnt]
		str.Format(L"%d,%s,%s,%s,%d,%d,%d,%s",
			info.m_nNumber, info.m_szRegion, info.m_szFlag, info.m_szName,
			info.m_bWoman ? 1 : 0, info.m_nRace, info.m_nPos, info.m_szTCPName);
		fputws(str, file);
		for (int i = 0; i < cTryCnt; i++) {
			str.Format(L",%s,%s,%s,%d,%s", info.m_szSpeed[i], info.m_szDistance[i], info.m_szWind[i], info.m_nRank[i], Score2Text(info.m_nScore[i]));
			fputws(str, file);
		}
		for (int i = 0; i < cTryCnt*cRefereeCnt; i++) {
			str.Format(L",%s", info.m_szReferee[i]);
			fputws(str, file);
		}
		fputws(L"\n", file);
	}
	fclose(file);
	return true;
}
//
bool CSLTramplinWatcherDlg::LoadProject(CString szPath, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"rt,ccs=UNICODE") != 0) {
		szError.Format(_T("Can't open file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	m_szProjectPath = szPath;
	// read players
	m_Players.clear();
	m_pLastInfo = nullptr;
	m_pCurrentInfo = nullptr;
	wchar_t tmp[4096];
	while (fgetws(tmp, 4096, file) != nullptr) {
		wchar_t *ptr = tmp;
		// nNumber
		wchar_t *szNumber = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// sRegion
		wchar_t *szRegion = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// sFlag
		wchar_t *szFlag = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// sName
		wchar_t *szName = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// bWoman
		wchar_t *szWoman = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// nRace
		wchar_t *szRace = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// nPos
		wchar_t *szPos = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// sTCPName
		wchar_t *szTCPName = ptr;
		if (!ScanData(ptr)) { ASSERT(false);  continue; }
		// sSpeed[cTryCnt] & sDistance[cTryCnt] & nRank[cTryCnt] & sScore[cTryCnt]
		wchar_t *szSpeed[cTryCnt];
		wchar_t *szDistance[cTryCnt];
		wchar_t *szWind[cTryCnt];
		wchar_t *szRank[cTryCnt];
		wchar_t *szScore[cTryCnt];
		for (int i = 0; i < cTryCnt; i++) {
			szSpeed[i] = ptr;
			if (!ScanData(ptr)) { ASSERT(false);  continue; }
			szDistance[i] = ptr;
			if (!ScanData(ptr)) { ASSERT(false);  continue; }
			szWind[i] = ptr;
			if (!ScanData(ptr)) { ASSERT(false);  continue; }
			szRank[i] = ptr;
			if (!ScanData(ptr)) { ASSERT(false);  continue; }
			szScore[i] = ptr;
			if (!ScanData(ptr)) { ASSERT(false);  continue; }
		}
		// sReferee[cTryCnt*cRefereeCnt]
		wchar_t *szReferee[cTryCnt*cRefereeCnt];
		for (int i = 0; i < cTryCnt*cRefereeCnt; i++) {
			szReferee[i] = ptr;
			if (!ScanData(ptr)) { ASSERT(false);  continue; }
		}
		// add player
		sPlayerInfo info;
		info.m_nNumber = _wtoi(szNumber);
		// nNumber, sRegion, sFlag, sName, bWoman, nRace, nPos,  sTCPName, nRank, sScore, sSpeed[cTryCnt], sDistance[cTryCnt], sReferee[cTryCnt*cRefereeCnt]
		info.m_szRegion = szRegion;
		info.m_szFlag = szFlag;
		info.m_szName = szName;
		info.m_bWoman = _wtoi(szWoman) != 0;
		info.m_nRace = _wtoi(szRace);
		info.m_nPos = _wtoi(szPos);
		info.m_szTCPName = szTCPName;
		for (int i = 0; i < cTryCnt; i++) {
			info.m_szSpeed[i] = szSpeed[i];
			info.m_szDistance[i] = szDistance[i];
			info.m_szWind[i] = szWind[i];
			info.m_nRank[i] = _wtoi(szRank[i]);
			info.m_nScore[i] = Text2Score(szScore[i]);
		}
		for (int i = 0; i < cTryCnt*cRefereeCnt; i++) {
			info.m_szReferee[i] = szReferee[i];
		}
		m_Players.push_back(info);
	}
	fclose(file);
	OnCbnSelchangeRace();
	return true;
}
//

inline bool ScanCsvData(char* &ptr, char* &data)
{
	char nEndChar = 0;
	if (*ptr == '\'' || *ptr == '\"')  nEndChar = *ptr++;
	data = ptr;
	while (*ptr != 0) {
		if (*ptr == '\n' || *ptr == '\r') {
			*ptr++ = 0;
			return true;
		}
		else if (*ptr == nEndChar) {
			nEndChar = 0;
			*ptr = 0;
		}
		else if (*ptr == ',' || *ptr == ';' || *ptr == '\t') {
			*ptr++ = 0;
			return true;
		}
		ptr++;
	}
	return false;
}
//
inline bool CheckName(CStringA text, const char *name)
{
	int pos = text.Find(name);
	if (pos < 0)  return false;
	if (pos > 0 && text[pos - 1] != ':')  return false;
	if (text[pos + strlen(name)] != 0)  return false;
	return true;
}
//
bool CSLTramplinWatcherDlg::ImportPlayers(CString szPath, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, szPath, L"rt") != 0) {
		szError.Format(_T("Can't open file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	// file exist -> reset local info & fill it from the file
	m_szPlayersPath = szPath;
	// scan caption
	char tmp[4096];
	if (fgets(tmp, 4096, file) == nullptr) {
		szError.Format(_T("Can't read file '%s'\r\nError = %d."), szPath, errno);
		return false;
	}
	int nID[24];
	for (int i = 0; i < _countof(nID); i++)  nID[i] = -1;
	char *ptr = tmp;
	for (int i = 0; i < _countof(nID); i++) {
		char *name;
		if (!ScanCsvData(ptr, name))  break;
		if (CheckName(name, "Number"))  nID[i] = eNumberColumn;
		else if (CheckName(name, "Region"))  nID[i] = eRegionColumn;
		else if(CheckName(name, "Name"))  nID[i] = eNameColumn;
		else if (CheckName(name, "Flag"))  nID[i] = eFlagCln;
		else if (CheckName(name, "Woman"))  nID[i] = eWomanCln;
		else if (CheckName(name, "Race"))  nID[i] = eRaceCln;
		else if (CheckName(name, "Pos"))  nID[i] = ePosColumn;
		else  nID[i] = -1;
	}
	// scan players
	m_Players.clear();
	m_pLastInfo = nullptr;
	m_pCurrentInfo = nullptr;
	while (fgets(tmp, 4096, file) != nullptr) {
		char *ptr = tmp;
		char *szNumber = tmp, *szRegion = tmp, *szName = tmp, *szFlag = tmp, *szWoman = tmp, *szRace = tmp, *szPos = tmp;
		int cnt = 0;
		for (int i = 0; i < _countof(nID); i++) {
			char *name;
			if (!ScanCsvData(ptr, name))  break;
			switch (nID[i]) {
				case eNumberColumn:	szNumber = name;	cnt++;	break;
				case eRegionColumn:	szRegion = name;	cnt++;	break;
				case eNameColumn:	szName = name;		cnt++;	break;
				case eFlagCln:		szFlag = name;		cnt++;	break;
				case eWomanCln:		szWoman = name;		cnt++;	break;
				case eRaceCln:		szRace = name;		cnt++;	break;
				case ePosColumn:	szPos = name;		cnt++;	break;
			}
		}
		if (cnt == 7) {
			// add player
			sPlayerInfo info;
			info.m_nNumber = atoi(szNumber);
			info.m_szRegion = szRegion;
			info.m_szFlag = szFlag;
			info.m_szName = szName;
			info.m_bWoman = atoi(szWoman) != 0;
			info.m_nRace = atoi(szRace);
			info.m_nPos = atoi(szPos);
			for (int i = 0; i < cTryCnt; i++) {
				info.m_nRank[i] = 0;
				info.m_nScore[i] = 0;
			}
			m_Players.push_back(info);
		}
	}
	fclose(file);
	// fill player's info
	OnCbnSelchangeRace();
	return true;
}
//
void CSLTramplinWatcherDlg::FillPlayerList()
{
	TCHAR szPos[10], szNumber[10], szRank[20];
	int nRaceId = m_Races[m_RaceList.GetCurSel()].m_nRace;
	int nRunkIndex = m_Run2.GetCheck() == BST_CHECKED ? 1 : 0;
	m_PlayerList.DeleteAllItems();
	for (auto info : m_Players) if (info.m_nRace == nRaceId) {
		_itot_s(info.m_nPos, szPos, 10);
		_itot_s(info.m_nNumber, szNumber, 10);
		_itot_s(info.m_nRank[nRunkIndex], szRank, 10);
		int n = m_PlayerList.InsertItem(m_PlayerList.GetItemCount(), _T(""));
		m_PlayerList.SetItemText(n, ePosColumn, szPos);
		m_PlayerList.SetItemText(n, eRankColumn, szRank);
		m_PlayerList.SetItemText(n, eNumberColumn, szNumber);
		m_PlayerList.SetItemText(n, eNameColumn, info.m_szName);
		m_PlayerList.SetItemText(n, eScoreColumn, Score2Text(info.m_nScore[nRunkIndex]));
		m_PlayerList.SetItemText(n, eRegionColumn, info.m_szRegion);
	}
}
//
//
void CSLTramplinWatcherDlg::OnStartTCPWatch()
{
	if (m_bWachig)  return;
	CString szError;
	if (StartWatching(m_nIP, m_nPort, szError)) {
		MessageBox(_T("Start TCP Watching"));
		if (m_WatchTCP.GetCheck() != BST_CHECKED)  m_WatchTCP.SetCheck(BST_CHECKED);
	} else {
		MessageBox(szError, _T("Start TCP Watching"), MB_ICONERROR);
		OnStopTCPWatch();
	}
}
//
void CSLTramplinWatcherDlg::OnStopTCPWatch()
{
	StopWatching();
	if (m_WatchTCP.GetCheck() != BST_UNCHECKED)  m_WatchTCP.SetCheck(BST_UNCHECKED);
}
//
bool CSLTramplinWatcherDlg::StartWatching(DWORD nAddress, DWORD nPort, CString &szError)
{
	StopWatching();
	SOCKET sNewSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (INVALID_SOCKET == sNewSocket) {
		szError.Format(_T("Can't create socket!\r\nError = %d."), WSAGetLastError());
		return false;
	}
	sockaddr_in server_addr;
	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons((u_short)nPort);
	//TCHAR szAddress[] = _T("127.0.0.1");
	//if (InetPton(AF_INET, szAddress, &server_addr.sin_addr) != TRUE) {
	//	szError.Format(_T("Wrang IP address!\r\nError = %d."), WSAGetLastError());
	//	closesocket(sNewSocket);
	//	return false;
	//}
	server_addr.sin_addr.s_addr = htonl(nAddress);
	if (connect(sNewSocket, (SOCKADDR*)&server_addr, (int)sizeof(server_addr)) == SOCKET_ERROR) {
		szError.Format(_T("Can't connect to specified IP & port!\r\nError = %d."), WSAGetLastError());
		closesocket(sNewSocket);
		return false;
	}
	m_sSocket = sNewSocket;
	m_bUpdateConnectionInfo = true;
	m_nTotalDataSize = 0;
	m_nDataSize = 0;
	m_eTCPDataType = eTCPUnknown;
	m_bWachig = true;
	m_TCPThread = std::thread([=] { TCPWork(); });
	return true;
}
//
void CSLTramplinWatcherDlg::StopWatching()
{
	m_bWachig = false;
	if (INVALID_SOCKET != m_sSocket) {
		closesocket(m_sSocket);
		m_sSocket = INVALID_SOCKET;
		m_bUpdateConnectionInfo = true;
		m_nTotalDataSize = 0;
	}
	if (m_TCPThread.joinable())
		//	m_TCPThread.join();
		m_TCPThread.detach();
}
//
void CSLTramplinWatcherDlg::TCPWork()
{
	char data[16 * 1024];
	while (m_bWachig && CheckTCP(data, sizeof(data))) {}
}
//
bool CSLTramplinWatcherDlg::CheckTCP(char *data, int size)
{
	if (INVALID_SOCKET == m_sSocket)  return false;
	int ln = recv(m_sSocket, data, size, 0);
	return SendMessage(WM_TCP_DATA, ln, (LPARAM)data) == TRUE;
}
//
//
bool CSLTramplinWatcherDlg::SaveStartTables(CString &szError)
{
	int nRaceId = m_Races[m_RaceList.GetCurSel()].m_nRace;
	int nRunIndex = m_Run2.GetCheck() == BST_CHECKED ? 1 : 0;
	std::vector<sPlayerInfo> players;
	if (0 == nRunIndex) { // sort by 'pos'
		for (const sPlayerInfo &info : m_Players) if (info.m_nRace == nRaceId)
			players.push_back(info);
		std::sort(players.begin(), players.end(),
			[](const sPlayerInfo &a, const sPlayerInfo &b) { return a.m_nPos < b.m_nPos; });
	}
	else { // sort by score1 (backward)
		for (const sPlayerInfo &info : m_Players) if (info.m_nRace == nRaceId && info.m_nRank[0] > 0)
			players.push_back(info);
		std::sort(players.begin(), players.end(),
			[](const sPlayerInfo &a, const sPlayerInfo &b) { return a.m_nRank[0] > b.m_nRank[0]; });
	}
	return SaveTables(players, _T("Start"), nRunIndex, szError);
}
//
bool CSLTramplinWatcherDlg::SaveResultTables(CString &szError)
{
	int nRaceId = m_Races[m_RaceList.GetCurSel()].m_nRace;
	int nRunIndex = m_Run2.GetCheck() == BST_CHECKED ? 1 : 0;
	std::vector<sPlayerInfo> players;
	for (const sPlayerInfo &info : m_Players) if (info.m_nRace == nRaceId && info.m_nRank[nRunIndex] > 0)
		players.push_back(info);
	std::sort(players.begin(), players.end(),
		[nRunIndex](const sPlayerInfo &a, const sPlayerInfo &b)	{ return a.m_nRank[nRunIndex] < b.m_nRank[nRunIndex]; });
	SetDlgItemInt(IDC_TABLE_CNT, (players.size() + cTableSize - 1) / cTableSize);
	return SaveTables(players, _T("Result"), nRunIndex, szError);
}
//
bool CSLTramplinWatcherDlg::SaveTables(std::vector<sPlayerInfo> players, CString szName, int nRunIndex, CString &szError)
{
	CPath path(m_szResultPath);
	CString dir = m_szResultPath.Left(path.FindFileName());
	CString tmp[cMaxTableCnt + 1], fname[cMaxTableCnt + 1];
	tmp[0].Format(_T("%stmp0.csv"), dir);
	fname[0].Format(_T("%s%s%d.csv"), dir, szName, nRunIndex + 1);
	for (int i = 1; i <= cMaxTableCnt; i++) {
		tmp[i].Format(_T("%stmp%d.csv"), dir, i);
		fname[i].Format(_T("%s%s%dp%d.csv"), dir, szName, nRunIndex + 1, i);
	}
	if (!SaveTablePart(players, 0, players.size(), tmp[0], nRunIndex, szError))  return false;
	for (int i = 0; i < cMaxTableCnt; i++) {
		if (!SaveTablePart(players, i*cTableSize, cTableSize, tmp[i + 1], nRunIndex, szError))  return false;
	}
	for (int i = 0; i <= cMaxTableCnt; i++) {
		for (int j = 0; j < 5; j++) {
			DeleteFile(fname[i]);
			if (MoveFile(tmp[i], fname[i]))  break;
			else  Sleep(100);
		}
	}
	return true;
}
//
bool CSLTramplinWatcherDlg::SaveTablePart(std::vector<sPlayerInfo> players, int nFrom, int nCnt, CString fname, int nRunIndex, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, fname, L"wt,ccs=UNICODE") != 0) {
		szError.Format(_T("Can't create file '%s'\r\nError = %d."), fname, errno);
		return false;
	}
	int nRB = nRunIndex * cRefereeCnt;
	// save header
	CStringW str;
	fputws(L"Pos,Rank,BIP,Name,Region,Flag,Speed,Distance,Score1,Score2,Score,Referee1,Referee2,Referee3,Referee4,Referee5\n", file);
	for (int i = 0; i < nCnt; i++) {
		int nIndex = nFrom + i;
		if(nIndex < (int)players.size()) {
			const sPlayerInfo &info = players[nIndex];
			int nScore[3];
			nScore[0] = info.m_nScore[0];
			if (info.m_nScore[1] > 0) {
				nScore[2] = info.m_nScore[1];
				nScore[1] = info.m_nScore[1] - info.m_nScore[0];
			}
			else {
				nScore[1] = 0;
				nScore[2] = info.m_nScore[0];
			}
			// Pos,Rank,BIP,Name,Region,Flag,Speed,Distance,Score1,Score2,Score,Referee1,Referee2,Referee3,Referee4,Referee5
			str.Format(L"%d,%d,%d,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s\n",
				nIndex + 1, info.m_nRank[nRunIndex], info.m_nNumber, info.m_szName, info.m_szRegion, info.m_szFlag,
				info.m_szSpeed[nRunIndex], info.m_szDistance[nRunIndex], Score2Text(nScore[0]), Score2Text(nScore[1]), Score2Text(nScore[2]),
				info.m_szReferee[nRB + 0], info.m_szReferee[nRB + 1], info.m_szReferee[nRB + 2], info.m_szReferee[nRB + 3], info.m_szReferee[nRB + 3]);
			fputws(str, file);
		}
		else  fputws(L"\n", file);
	}
	fclose(file);
	return true;
}
//
//
bool CSLTramplinWatcherDlg::SavePlayer(sPlayerInfo *info, bool bLast, CString &szError)
{
	CPath path(m_szResultPath);
	CString dir = m_szResultPath.Left(path.FindFileName());
	CString tmp, fname;
	tmp.Format(_T("%stmp.csv"), dir);
	fname.Format(_T("%s%s.csv"), dir, bLast ? _T("Last") : _T("Current"));
	int nRunIndex = m_Run2.GetCheck() == BST_CHECKED ? 1 : 0;
	if (!SavePlayerLocal(info, tmp, nRunIndex, szError))  return false;
	for (int j = 0; j < 10; j++) {
		DeleteFile(fname);
		if (MoveFile(tmp, fname))  break;
		else  Sleep(100);
	}
	return true;
}
//
bool CSLTramplinWatcherDlg::SavePlayerLocal(sPlayerInfo *info, CString fname, int nRunIndex, CString &szError)
{
	FILE *file = nullptr;
	if (_wfopen_s(&file, fname, L"wt,ccs=UNICODE") != 0) {
		szError.Format(_T("Can't create file '%s'\r\nError = %d."), fname, errno);
		return false;
	}
	int nRB = nRunIndex * cRefereeCnt;
	// save header
	CStringW str;
	fputws(L"Pos,Rank,BIP,Name,Region,Flag,Speed,Distance,Wind,Score1,Score2,Score,Referee1,Referee2,Referee3,Referee4,Referee5\n", file);
	// save info
	int nScore[3];
	nScore[0] = info->m_nScore[0];
	if (info->m_nScore[1] > 0) {
		nScore[2] = info->m_nScore[1];
		nScore[1] = info->m_nScore[1] - info->m_nScore[0];
	}
	else {
		nScore[1] = 0;
		nScore[2] = info->m_nScore[0];
	}
	// Pos,Rank,BIP,Name,Region,Flag,Speed,Distance,Wind,Score1,Score2,Score,Referee1,Referee2,Referee3,Referee4,Referee5
	str.Format(L"%d,%d,%d,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s\n",
		info->m_nPos, info->m_nRank[nRunIndex], info->m_nNumber, info->m_szName, info->m_szRegion, info->m_szFlag,
		info->m_szSpeed[nRunIndex], info->m_szDistance[nRunIndex], info->m_szWind[nRunIndex], Score2Text(nScore[0]), Score2Text(nScore[1]), Score2Text(nScore[2]),
		info->m_szReferee[nRB + 0], info->m_szReferee[nRB + 1], info->m_szReferee[nRB + 2], info->m_szReferee[nRB + 3], info->m_szReferee[nRB + 4]);
	fputws(str, file);
	fclose(file);
	return true;
}
//
bool CSLTramplinWatcherDlg::CurrentIsReady()
{
	if (nullptr == m_pCurrentInfo)  return false;
	int nRunIndex = m_Run2.GetCheck() == BST_CHECKED ? 1 : 0;
	int nRB = nRunIndex * cRefereeCnt;
	return	!m_pCurrentInfo->m_szReferee[nRB + 0].IsEmpty() &&
			!m_pCurrentInfo->m_szReferee[nRB + 1].IsEmpty() &&
			!m_pCurrentInfo->m_szReferee[nRB + 2].IsEmpty() &&
			!m_pCurrentInfo->m_szReferee[nRB + 3].IsEmpty() &&
			!m_pCurrentInfo->m_szReferee[nRB + 3].IsEmpty();
}
//
//
void CSLTramplinWatcherDlg::ParseData(char *data, int size)
{
	int nPos = 0;
	if (m_nDataSize > 0) {
		int rest = _countof(m_szData) - m_nDataSize - 1;
		while (nPos < rest && nPos < size) {
			if (data[nPos] != ';')  m_szData[m_nDataSize++] = data[nPos++];
			else {
				nPos++;
				m_szData[m_nDataSize] = 0;
				ParseCommand();
				m_nDataSize = 0;
				break;
			}
		}
	}
	while (nPos < size) {
		if (data[nPos] == '$') // start new command
			m_szData[m_nDataSize++] = data[nPos++];
		else if (m_nDataSize > 0) {
			if (data[nPos] != ';')  m_szData[m_nDataSize++] = data[nPos++];
			else {
				nPos++;
				m_szData[m_nDataSize] = 0;
				ParseCommand();
				m_nDataSize = 0;
			}
		} else { // skip symbol
			nPos++;
		}
	}
}
//
inline bool CompareCommand(const WCHAR *text, const WCHAR *cmd, CString &param)
{
	int i;
	for (i = 0; cmd[i] != 0; i++)
		if (text[i] != cmd[i])  return false;
	param = CString(&text[i]).Trim();
CString sss;  sss.Format(_T("TRAM_CMD:%s\r\n"), text);	OutputDebugString(sss);
	return true;
}
//
inline bool UpdateTextParam(CString &str, CString param)
{
	if (!param.IsEmpty()) {
		str = param;
		return true;
	}
	return false;
}
//
void CSLTramplinWatcherDlg::ParseCommand()
{
	std::wstring_convert<std::codecvt_utf8<wchar_t>, wchar_t> utf8conv;
	std::wstring wstr = utf8conv.from_bytes(&m_szData[1]);
	{ CString str(wstr.c_str());	OutputDebugString(str + _T("\r\n")); }
	const WCHAR *text = (const WCHAR*)wstr.c_str();
	// parse string
	CString param;
	if (CompareCommand(text, L"ClearList", param)) { // start list
		if (eTCPCurrent == m_eTCPDataType  &&  nullptr != m_pCurrentInfo) {
			m_MarkCurrent.ShowWindow(TRUE);
			SetTimer(cTimer4LastPlayer, cTimeout4LastPlayer, NULL);
			m_pLastInfo = m_pCurrentInfo;
		}
		// 
		m_eTCPDataType = eTCPList;
		m_nLastRunIndex = m_Run2.GetCheck() == BST_CHECKED ? 1 : 0;
		m_pCurrentInfo = nullptr;
		m_bUpdateCurrentInfo = true;
		m_nLastRank = 0;
		return;
	}
	if (CompareCommand(text, L"BibCurrent:", param)) { // start new player
		m_eTCPDataType = eTCPCurrent;
		m_nLastRunIndex = m_Run2.GetCheck() == BST_CHECKED ? 1 : 0;
		int nNumber = _ttoi(&wstr.c_str()[11]);
		m_pCurrentInfo = GetPlayer(nNumber);
		m_bUpdateCurrentInfo = true;
		return;
	}
	if (eTCPCurrent == m_eTCPDataType) { // -------------------- Current Player Info --------------
		if (m_pCurrentInfo == nullptr)  return;
		else if (CompareCommand(text, L"NameCurrent:", param)) {
			m_pCurrentInfo->m_szTCPName = param;
		}
		else if (CompareCommand(text, L"R1SpeedCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szSpeed[0], param)  &&  0 == m_nLastRunIndex)
				SetCurrentSpeed(param);
		}
		else if (CompareCommand(text, L"R2SpeedCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szSpeed[1], param)  &&  1 == m_nLastRunIndex)
				SetCurrentSpeed(param);
		}
		else if (CompareCommand(text, L"R1DistanceCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szDistance[0], param)  &&  0 == m_nLastRunIndex)
				SetCurrentDistance(param);
		}
		else if (CompareCommand(text, L"R2DistanceCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szDistance[1], param)  &&  1 == m_nLastRunIndex)
				SetCurrentDistance(param);
		}
		else if (CompareCommand(text, L"R1WindPointCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szWind[0], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R2WindPointCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szWind[1], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R1ACurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[0 * cRefereeCnt + 0], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R1BCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[0 * cRefereeCnt + 1], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R1CCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[0 * cRefereeCnt + 2], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R1DCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[0 * cRefereeCnt + 3], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R1ECurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[0 * cRefereeCnt + 4], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R2ACurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[1 * cRefereeCnt + 0], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R2BCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[1 * cRefereeCnt + 1], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R2CCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[1 * cRefereeCnt + 2], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R2DCurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[1 * cRefereeCnt + 3], param))
				m_bCheckCurrentInfoFullness = true;
		}
		else if (CompareCommand(text, L"R2ECurrent:", param)) {
			if (UpdateTextParam(m_pCurrentInfo->m_szReferee[1 * cRefereeCnt + 4], param))
				m_bCheckCurrentInfoFullness = true;
		}
		m_bUpdateCurrentInfo = true;
	}
	else if (eTCPList == m_eTCPDataType) { // -------------------- List Info ---------------
		if (CompareCommand(text, L"Rank:", param)) {
			m_bRedrawPlayersInfo = true;
			m_nLastRank = _ttoi(param);
			return;
		}
		if (CompareCommand(text, L"Name:", param)) {
			m_pCurrentInfo = FindPlayer(param);
			if (nullptr == m_pCurrentInfo)  return;
			m_pCurrentInfo->m_nRank[m_nLastRunIndex] = m_nLastRank;
			m_nLastRank = 0;
		}
		if (nullptr == m_pCurrentInfo)  return;
		if (CompareCommand(text, L"R2Distance:", param)) {
			UpdateTextParam(m_pCurrentInfo->m_szDistance[0], param);
		}
		else if (CompareCommand(text, L"R3Distance:", param)) {
			UpdateTextParam(m_pCurrentInfo->m_szDistance[1], param);
		}
		else if (CompareCommand(text, L"Points:", param)) {
			m_pCurrentInfo->m_nScore[m_nLastRunIndex] = Text2Score(param);
		}
		m_bUpdateCurrentInfo = true;
	}
}
//
sPlayerInfo *CSLTramplinWatcherDlg::GetPlayer(int nNumber)
{
	int nRace = m_Races[m_RaceList.GetCurSel()].m_nRace;
	for (sPlayerInfo & info : m_Players) if (info.m_nNumber == nNumber && info.m_nRace == nRace)
		return &info;
	return nullptr;
}
//
sPlayerInfo *CSLTramplinWatcherDlg::FindPlayer(const WCHAR *name)
{
	int nRace = m_Races[m_RaceList.GetCurSel()].m_nRace;
	for (sPlayerInfo & info : m_Players) if (info.m_nRace == nRace) {
		int ln = info.m_szTCPName.GetLength();
		if (ln > 0 && _tcsnccmp(info.m_szTCPName, name, ln) == 0)
			return &info;
	}
	return nullptr;
}
//
void CSLTramplinWatcherDlg::SetCurrentSpeed(CString szValue)
{
	if (eTCPCurrent != m_eTCPDataType || nullptr == m_pCurrentInfo)  return;
	double fValue = _ttof(szValue);
	if (fValue < 1.0)  return;
	CString szError;
	if (SavePlayer(m_pCurrentInfo, false, szError)) {
		SendSLTitlerCmd(cTitlerCmd4SpeedOn);
		SetTimer(cTimer4SpeedOff, cTimeout4SpeedOff, NULL);
	}
}
//
void CSLTramplinWatcherDlg::SetCurrentDistance(CString szValue)
{
	if (eTCPCurrent != m_eTCPDataType || nullptr == m_pCurrentInfo)  return;
	double fValue = _ttof(szValue);
	if (fValue < 1.0)  return;
	CString szError;
	if (SavePlayer(m_pCurrentInfo, false, szError)) {
		SendSLTitlerCmd(cTitlerCmd4DistanceOn);
		SetTimer(cTimer4DistanceOff, cTimeout4DistanceOff, NULL);
	}
}
//
BOOL CALLBACK SendTitleMessage(HWND hwnd, LPARAM lParam)
{
	static const TCHAR szCaption[] = _T("SLSportTitler");
	static const int lnCaption = _countof(szCaption) - 1;
	TCHAR name[MAX_PATH];
	int ln = GetWindowText(hwnd, name, MAX_PATH);
	if (ln >= lnCaption && _tcsncicmp(name, szCaption, lnCaption) == 0) {
		::PostMessage(hwnd, WM_HOTKEY, 0, lParam);
		return FALSE;
	}
	return TRUE;
}
//
void CSLTramplinWatcherDlg::SendSLTitlerCmd(int nCmd)
{
	EnumWindows(SendTitleMessage, MAKELONG(0, nCmd));
}