
// SLTramplinWatcherDlg.h : header file
//
#pragma once

#include "afxcmn.h"
#include "afxwin.h"
#include <list>
#include <vector>
#include <thread>

const int cTryCnt = 2;
const int cRefereeCnt = 5;
const int cMaxTableCnt = 5;
const int cTableSize = 10;

struct sPlayerInfo {
	// unique ID = BIP = Number
	int m_nNumber;
	// SLSportEditor info
	CString m_szRegion;
	CString m_szFlag;
	CString m_szName;
	int m_bWoman;
	int m_nRace;
	int m_nPos;
	// TCP info
	CString m_szTCPName;
	CString m_szSpeed[cTryCnt];
	CString m_szDistance[cTryCnt];
	CString m_szWind[cTryCnt];
	int m_nRank[cTryCnt];
	int m_nScore[cTryCnt];
	CString m_szReferee[cTryCnt*cRefereeCnt];
};

inline int Text2Score(CString text)
{
	double v = _wtof(text);
	int n = (int)(v * 10 + 0.499);
	return n;
}
inline CString Score2Text(int score)
{
	CString str;
	str.Format(_T("%d.%d"), score / 10, score % 10);
	return str;
}


struct sRaceInfo {
	int m_nRace;
	bool m_bWoman;
	CString m_szName;
};


// CSLTramplinWatcherDlg dialog
class CSLTramplinWatcherDlg : public CDialogEx
{
// Construction
public:
	CSLTramplinWatcherDlg(CWnd* pParent = NULL);	// standard constructor
	virtual ~CSLTramplinWatcherDlg() { StopWatching(); }
	void SaveSettings(CWinApp *pApp);
	void LoadSettings(CWinApp *pApp);

// Dialog Data
#ifdef AFX_DESIGN_TIME
	enum { IDD = IDD_SLTRAMPLINWATCHER_DIALOG };
#endif

	protected:
	virtual void DoDataExchange(CDataExchange* pDX);	// DDX/DDV support


// Implementation
protected:
	HICON m_hIcon;
	CButton m_WatchTCP;
	CComboBox m_RaceList;
	CButton m_Run2;
	CListCtrl m_PlayerList;
	CStatic m_TableCnt;
	CStatic m_TCPInfo;
	CStatic m_CurrentInfo;
	CStatic m_MarkCurrent;
	//
	CString m_szProjectPath;
	CString m_szPlayersPath;
	CString m_szResultPath;
	DWORD m_nIP;
	DWORD m_nPort;
	// TCP data
	volatile bool m_bWachig;
	SOCKET m_sSocket;
	std::thread m_TCPThread;
	char m_szData[1024];
	int m_nDataSize;
	int m_nTotalDataSize;
	enum { eTCPUnknown=0, eTCPCurrent, eTCPList } m_eTCPDataType;
	int m_nLastRank;
	int m_nLastRunIndex;
	sPlayerInfo *m_pCurrentInfo;
	sPlayerInfo *m_pLastInfo;
	bool m_bUpdateConnectionInfo;
	bool m_bUpdateCurrentInfo;
	bool m_bCheckCurrentInfoFullness;
	bool m_bRedrawPlayersInfo;
	// DB data
	enum { ePosColumn = 0, eRankColumn, eNumberColumn, eNameColumn, eRegionColumn, eScoreColumn, eLastColumn,
			eWomanCln, eRaceCln, eFlagCln };
	std::vector<sRaceInfo> m_Races;
	std::list<sPlayerInfo> m_Players;
protected:
	bool LoadRaces();
	bool SaveProject(CString szPath, CString &szError);
	bool LoadProject(CString szPath, CString &szError);
	bool ImportPlayers(CString szPath, CString &szError);
	void FillPlayerList();
	void OnStartTCPWatch();
	void OnStopTCPWatch();
	bool StartWatching(DWORD nAddress, DWORD nPort, CString &szError);
	void StopWatching();
	bool CheckTCP(char *data, int size);
	bool SaveStartTables(CString &szError);
	bool SaveResultTables(CString &szError);
	bool SaveTables(std::vector<sPlayerInfo> players, CString szName, int nRunIndex, CString &szError);
	bool SaveTablePart(std::vector<sPlayerInfo> players, int nFrom, int nCnt, CString fname, int nRunIndex, CString &szError);
	bool SavePlayer(sPlayerInfo *info, bool bLast, CString &szError);
	bool SavePlayerLocal(sPlayerInfo *info, CString fname, int nRunIndex, CString &szError);
	bool CurrentIsReady();
	void TCPWork();
	void ParseData(char *data, int size);
	void ParseCommand();
	sPlayerInfo *GetPlayer(int nNumber);
	sPlayerInfo *FindPlayer(const WCHAR *nNumber);
	void SetCurrentSpeed(CString szValue);
	void SetCurrentDistance(CString szValue);
	void SendSLTitlerCmd(int nCmd);

public:
	// Generated message map functions
	virtual BOOL OnInitDialog();
	virtual void OnOK() {}
	virtual void OnCancel() {}
	afx_msg void OnSysCommand(UINT nID, LPARAM lParam);
	afx_msg void OnPaint();
	afx_msg HCURSOR OnQueryDragIcon();
	DECLARE_MESSAGE_MAP()
	afx_msg void OnBnClickedExit();
	afx_msg void OnBnClickedImport();
	afx_msg void OnBnClickedSaveProject();
	afx_msg void OnBnClickedLoadProject();
	afx_msg void OnBnClickedTcpSettings();
	afx_msg void OnBnClickedTcpWatch();
	afx_msg void OnCbnSelchangeRace();
	afx_msg void OnBnClickedResultPath();
	afx_msg void OnBnClickedMakeStart();
	afx_msg void OnBnClickedMakeResult();
	afx_msg void OnBnClickedMakeLast();
	afx_msg void OnBnClickedRun2();
	afx_msg void OnNMDblclkPlayers(NMHDR *pNMHDR, LRESULT *pResult);
	afx_msg void OnTimer(UINT_PTR nIDEvent);
	afx_msg LRESULT OnTCPData(WPARAM, LPARAM);
};
