
// SLVolaSplitterDlg.cpp : implementation file
//

#include "stdafx.h"
#include "SLVolaSplitter.h"
#include "SLVolaSplitterDlg.h"
#include "afxdialogex.h"
#include <thread>
#include "Ws2tcpip.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

const TCHAR gsCaption[] = _T("SLVolaSplitter");
const long WM_OPEN_CONNECTION = WM_USER + 1;
const long WM_CLOSE_CONNECTION = WM_USER + 2;

void StartConnectionThread(CSLVolaSplitterDlg *dlg)
{
	dlg->ConnectionThread();
}

inline void IPName2DWORD(const TCHAR *name, BYTE addr[4])
{
	int pos = 0;
	for (int i = 0; i < 4; i++) {
		BYTE n = 0;
		while (name[pos] >= _T('0') && name[pos] <= _T('9')) {
			n *= 10;
			n += name[pos++] - _T('0');
		}
		if (name[pos] != _T('\0'))  pos++;
		addr[i] = n;
	}
}

void SetIP(const TCHAR *name, CIPAddressCtrl &ctrl)
{
	BYTE addr[4];
	IPName2DWORD(name, addr);
	ctrl.SetAddress(addr[0], addr[1], addr[2], addr[3]);
}

void GetIP(CIPAddressCtrl &ctrl, CString &name)
{
	BYTE addr[4];
	ctrl.GetAddress(addr[0], addr[1], addr[2], addr[3]);
	name.Format(_T("%d.%d.%d.%d"), addr[0], addr[1], addr[2], addr[3]);
}

bool GoodIP(const TCHAR *name)
{
	BYTE addr[4];
	IPName2DWORD(name, addr);
	return addr[0] != 0 && addr[1] != 0 && addr[2] != 0 && addr[3] != 0;
}


// CSLVolaSplitterDlg dialog
CSLVolaSplitterDlg::CSLVolaSplitterDlg(DWORD nServerPort, CString szClientIP[cMaxClientCnt], DWORD nClientPort[cMaxClientCnt])
 : CDialogEx(IDD_SLVOLASPLITTER_DIALOG, NULL)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
	m_nServerPort = nServerPort;
	for (int i = 0; i < cMaxClientCnt; i++) {
		m_szClientIP[i] = szClientIP[i];
		m_nClientPort[i] = nClientPort[i];
		m_sClientSocket[i] = INVALID_SOCKET;
	}
	m_sVolaSocket = INVALID_SOCKET;
	m_sServerSocket = INVALID_SOCKET;
	m_hEvent4Connect = NULL;
	m_bRunning = false;
	m_bSuicide = false;
	m_pTCPDataHead = NULL;
	m_pTCPDataTail = NULL;
	m_nTCPPacketCnt = 0;
	m_nTCPDataSize = 0;
}

void CSLVolaSplitterDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialogEx::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_SERVER_PORT, m_ServerPort);
	DDX_Control(pDX, IDC_START_SERVER, m_ServerRun);
	DDX_Control(pDX, IDC_SERVER_STATE, m_ServerState);
	for (int i = 0; i < cMaxClientCnt; i++) {
		DDX_Control(pDX, IDC_IPADDRESS1 + i, m_IPAddress[i]);
		DDX_Control(pDX, IDC_IPPORT1 + i, m_IPPort[i]);
		DDX_Control(pDX, IDC_IPRUN1 + i, m_IPRun[i]);
		DDX_Control(pDX, IDC_IPSTATUS1 + i, m_IPState[i]);
	}
}

BEGIN_MESSAGE_MAP(CSLVolaSplitterDlg, CDialogEx)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_DESTROY()
	ON_WM_TIMER()
	ON_MESSAGE(WM_OPEN_CONNECTION, &CSLVolaSplitterDlg::OnOpenConnection)
	ON_BN_CLICKED(IDC_START_SERVER, &CSLVolaSplitterDlg::OnBnClickedRun)
	ON_BN_CLICKED(IDC_IPRUN1, &CSLVolaSplitterDlg::OnBnClickedIpRun1)
	ON_BN_CLICKED(IDC_IPRUN2, &CSLVolaSplitterDlg::OnBnClickedIpRun2)
	ON_BN_CLICKED(IDC_IPRUN3, &CSLVolaSplitterDlg::OnBnClickedIpRun3)
	ON_BN_CLICKED(IDC_IPRUN4, &CSLVolaSplitterDlg::OnBnClickedIpRun4)
	ON_BN_CLICKED(IDC_IPRUN5, &CSLVolaSplitterDlg::OnBnClickedIpRun5)
END_MESSAGE_MAP()

// CSLVolaSplitterDlg message handlers
BOOL CSLVolaSplitterDlg::OnInitDialog()
{
	CDialogEx::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		BOOL bNameValid;
		CString strAboutMenu;
		bNameValid = strAboutMenu.LoadString(IDS_ABOUTBOX);
		ASSERT(bNameValid);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	// init locals
	Locals2DlgItems();
	for (int i = 0; i < cMaxClientCnt; i++) if (GoodIP(m_szClientIP[i]))  m_IPRun[i].SetCheck(BST_CHECKED);

	CString szError;
	if (!InitServer(szError)) {
		MessageBox(szError, gsCaption, MB_ICONSTOP);
		EndDialog(IDCANCEL);
	}
	SetTimer(101, 10, NULL);
	//Test();

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CSLVolaSplitterDlg::OnDestroy()
{
	KillTimer(101);
	DlgItems2Locals();
	m_bSuicide = true;
	CDialogEx::OnDestroy();
	if (m_hEvent4Connect != NULL)  CloseHandle(m_hEvent4Connect);
	m_hEvent4Connect = NULL;
	StopAll();
	WSACleanup();
	ClearVolaData();
}

void CSLVolaSplitterDlg::Locals2DlgItems()
{
	SetDlgItemInt(IDC_SERVER_PORT, m_nServerPort, FALSE);
	for (int i = 0; i < cMaxClientCnt; i++) {
		SetIP(m_szClientIP[i], m_IPAddress[i]);
		SetDlgItemInt(IDC_IPPORT1 + i, m_nClientPort[i], FALSE);
	}
}

void CSLVolaSplitterDlg::DlgItems2Locals()
{
	m_nServerPort = GetDlgItemInt(IDC_SERVER_PORT);
	for (int i = 0; i < cMaxClientCnt; i++) {
		GetIP(m_IPAddress[i], m_szClientIP[i]);
		m_nClientPort[i] = GetDlgItemInt(IDC_IPPORT1 + i);
	}
}

void CSLVolaSplitterDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CDialogEx dlgAbout(IDD_ABOUTBOX);
		dlgAbout.DoModal();
	}
	else
	{
		CDialogEx::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CSLVolaSplitterDlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialogEx::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CSLVolaSplitterDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

void CSLVolaSplitterDlg::OnBnClickedRun()
{
	if (!m_bRunning) {
		CString szError;
		DlgItems2Locals();
		ClearVolaData();
		if (!StartServer(m_nServerPort, szError)) {
			MessageBox(szError, _T("Init Server (VOLA)"), MB_ICONSTOP);
			return;
		}
		m_bRunning = true;
		for (int i = 0; i < cMaxClientCnt; i++) {
			bool bRunClient = RunClient(i);
			BOOL nEnable = bRunClient ? FALSE : TRUE;
			m_IPAddress[i].EnableWindow(nEnable);
			m_IPPort[i].EnableWindow(nEnable);
		}
		m_ServerRun.SetWindowTextW(_T("Stop"));
		m_ServerPort.EnableWindow(FALSE);
	} else {
		StopAll();
		for (int i = 0; i < cMaxClientCnt; i++)  StopClient(i);
		m_bRunning = false;
		m_ServerRun.SetWindowTextW(_T("Start"));
		m_ServerPort.EnableWindow(TRUE);
		for (int i = 0; i < cMaxClientCnt; i++) {
			m_IPAddress[i].EnableWindow(TRUE);
			m_IPPort[i].EnableWindow(TRUE);
		}
	}
}

LRESULT CSLVolaSplitterDlg::OnOpenConnection(WPARAM, LPARAM)
{
	if (m_bSuicide)  return 0;
	sockaddr_in client_addr;
	int ln = sizeof(client_addr);
	SOCKET sNewSocket = accept(m_sServerSocket, (sockaddr*)&client_addr, &ln);
	if (sNewSocket == INVALID_SOCKET) {
		CString str;
		str.Format(_T("Can't accept TCP/IP client = VOLA server!\r\nERROR = %d"), WSAGetLastError());
		MessageBox(str, gsCaption, MB_ICONEXCLAMATION);
		return 0;
	}
	ClearVolaData();
	if (INVALID_SOCKET != m_sVolaSocket)  closesocket(m_sVolaSocket);
	m_sVolaSocket = sNewSocket;
	return 1;
}

void CSLVolaSplitterDlg::OnTimer(UINT_PTR nIDEvent)
{
	static bool bInTimer = false;
	if (!bInTimer) {
		bInTimer = true;
		if (m_bRunning) for (int i = 0; i < 100; ++i) if (!CheckVola())  break;
		UpdateVolaInfo();
		for (int i = 0; i < cMaxClientCnt; i++)  UpdateClientInfo(i);
		bInTimer = false;
	}
	CDialogEx::OnTimer(nIDEvent);
}

bool CSLVolaSplitterDlg::InitServer(CString &szError)
{
	// init socket library
	WSADATA wsaData;
	if (WSAStartup(MAKEWORD(2, 2), &wsaData) > 0) {
		szError.Format(_T("Can't init socket library!!!\r\nError = %d."), WSAGetLastError());
		return false;
	}
	m_hEvent4Connect = WSACreateEvent();
	if (NULL == m_hEvent4Connect) {
		szError.Format(_T("Can't create WSAEvent!!!\r\nError = %d."), WSAGetLastError());
		return false;
	}
	std::thread thWaitConnect(StartConnectionThread, this);
	thWaitConnect.detach();
	return true;
}

void CSLVolaSplitterDlg::StopAll()
{
	if (m_sServerSocket != INVALID_SOCKET)  closesocket(m_sServerSocket);
	m_sServerSocket = INVALID_SOCKET;
	if (m_sVolaSocket != INVALID_SOCKET) {
		closesocket(m_sVolaSocket);
		m_sVolaSocket = INVALID_SOCKET;
	}
	for (int i = 0; i < cMaxClientCnt; i++) if(m_sClientSocket[i] != INVALID_SOCKET) {
		closesocket(m_sClientSocket[i]);
		m_sClientSocket[i] = INVALID_SOCKET;
	}
}

void CSLVolaSplitterDlg::ConnectionThread()
{
	while (!m_bSuicide) {
		DWORD res = WSAWaitForMultipleEvents(1, &m_hEvent4Connect, FALSE, 1000, TRUE);
		if (WSA_WAIT_EVENT_0 == res) {
			PostMessage(WM_OPEN_CONNECTION, 0, 0);
			WSAResetEvent(m_hEvent4Connect);
		}
	}
}

bool CSLVolaSplitterDlg::StartServer(DWORD nPort, CString &szError)
{
	// init TCP connection
	SOCKET sNewSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (INVALID_SOCKET == sNewSocket) {
		int error = WSAGetLastError();
		szError.Format(_T("Can't create socket!\r\nERROR = %d."), error);
		return false;
	}
	sockaddr_in local_addr;
	local_addr.sin_family = AF_INET;
	local_addr.sin_port = htons((u_short)nPort);
	local_addr.sin_addr.s_addr = ADDR_ANY;
	if (bind(sNewSocket, (SOCKADDR*)&local_addr, (int)sizeof(local_addr)) == SOCKET_ERROR) {
		int error = WSAGetLastError();
		szError.Format(_T("Can't bind socket!\r\nERROR = %d."), error);
		closesocket(sNewSocket);
		return false;
	}
	if (listen(sNewSocket, SOMAXCONN) == SOCKET_ERROR) {
		int error = WSAGetLastError();
		szError.Format(_T("Can't listen socket!\r\nERROR = %d."), error);
		closesocket(sNewSocket);
		return false;
	}
	if (SOCKET_ERROR == WSAEventSelect(sNewSocket, m_hEvent4Connect, FD_ACCEPT)) {
		int error = WSAGetLastError();
		szError.Format(_T("Error on call WSAEventSelect!\r\nERROR = %d."), error);
		closesocket(sNewSocket);
		return false;
	}
	// all right
	m_sServerSocket = sNewSocket;
	return true;
}

void CSLVolaSplitterDlg::ClearVolaData()
{
	while (m_pTCPDataHead != NULL) {
		TCPData *ptr = m_pTCPDataHead;
		m_pTCPDataHead = m_pTCPDataHead->m_pNext;
		delete ptr;
	}
	m_pTCPDataTail = NULL;
	m_nTCPPacketCnt = 0;
	m_nTCPDataSize = 0;
}

bool CSLVolaSplitterDlg::CheckVola()
{
	if (INVALID_SOCKET == m_sVolaSocket)  return false;
	char data[4096];
	int ln = recv(m_sVolaSocket, data, sizeof(data), 0);
	if (ln > 0) { // add data
		if (AddVolaData(data, ln)) { // send packet to all clients
			for (int i = 0; i < cMaxClientCnt; i++)
				if (INVALID_SOCKET != m_sClientSocket[i])
					send(m_sClientSocket[i], data, ln, 0);
		}
		return true;
	} else if (ln == 0) {
		OnBnClickedRun();
		MessageBox(_T("VOLA stopped !!!"), gsCaption, MB_ICONEXCLAMATION);
	}
	return false;
}

void CSLVolaSplitterDlg::UpdateVolaInfo()
{
	CString str,tmp;
	const TCHAR *strRun = m_bRunning ? _T("Running") : _T("Stopped");
	const TCHAR *strTCP = INVALID_SOCKET == m_sVolaSocket ? _T("") : _T(" + connected");
	str.Format(_T("%s%s...\r\nPackets: Cnt=%d, Size=%d"), strRun, strTCP, m_nTCPPacketCnt, m_nTCPDataSize);
	m_ServerState.GetWindowText(tmp);
	if( tmp != str)
		m_ServerState.SetWindowText(str);
}

void CSLVolaSplitterDlg::ClickIPClient(int nIndex)
{
	bool bRunClient = RunClient(nIndex);
	BOOL nEnable = bRunClient ? FALSE : TRUE;
	m_IPAddress[nIndex].EnableWindow(nEnable);
	m_IPPort[nIndex].EnableWindow(nEnable);
}

bool CSLVolaSplitterDlg::RunClient(int nIndex)
{
	bool bChecked = BST_CHECKED == m_IPRun[nIndex].GetCheck();
	if (!m_bRunning)
		return bChecked;
	StopClient(nIndex);
	if (!bChecked)
		return false;
	m_IPState[nIndex].SetWindowText(_T("Init"));
	m_IPState[nIndex].UpdateWindow();
	CString szError;  SOCKET socket;
	if (!StartClient(nIndex, m_szClientIP[nIndex], m_nClientPort[nIndex], socket, szError)) {
		CString str;  str.Format(_T("Init client #%d"), nIndex + 1);
		MessageBox(szError, str, MB_ICONEXCLAMATION);
		m_IPRun[nIndex].SetCheck(BST_UNCHECKED);
		return false;
	}
	if (!SendVolaData(socket, szError)) {
		CString str;  str.Format(_T("Send data to client #%d"), nIndex + 1);
		MessageBox(szError, str, MB_ICONEXCLAMATION);
		m_IPRun[nIndex].SetCheck(BST_UNCHECKED);
		return false;
	}
	m_sClientSocket[nIndex] = socket;
	return true;
}

bool CSLVolaSplitterDlg::StartClient(int nIndex, CString szAddress, DWORD nPort, SOCKET& sSocket, CString &szError)
{
	SOCKET sNewSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (INVALID_SOCKET == sNewSocket) {
		szError.Format(_T("Can't create socket!\r\nError = %d."), WSAGetLastError());
		return false;
	}
	sockaddr_in server_addr;
	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons((u_short)nPort);
	if (InetPton(AF_INET, szAddress, &server_addr.sin_addr) != TRUE) {
		szError.Format(_T("Wrang IP address!\r\nError = %d."), WSAGetLastError());
		closesocket(sNewSocket);
		return false;
	}
	if (connect(sNewSocket, (SOCKADDR*)&server_addr, (int)sizeof(server_addr)) == SOCKET_ERROR) {
		szError.Format(_T("Can't connect to specified IP & port!\r\nError = %d."), WSAGetLastError());
		closesocket(sNewSocket);
		return false;
	}
	sSocket = sNewSocket;
	return true;
}

void CSLVolaSplitterDlg::StopClient(int nIndex)
{
	if (INVALID_SOCKET != m_sClientSocket[nIndex]) {
		closesocket(m_sClientSocket[nIndex]);
		m_sClientSocket[nIndex] = INVALID_SOCKET;
	}
}

bool CSLVolaSplitterDlg::SendVolaData(SOCKET sSocket, CString &szError)
{
	CWaitCursor wait;
	int nSentCnt = 0, nSendSize = 0, nErrorCnt = 0;
	for (TCPData *node = m_pTCPDataHead; node != NULL; ) {
		if ((nSentCnt + 1) % 100 == 0) // check packets from VOLA and send it to other clients
			CheckVola();
		int result = send(sSocket, node->m_pData, node->m_nSize, 0);
		if (SOCKET_ERROR == result) {
			int error = WSAGetLastError();
			if (WSAENOBUFS == error && nErrorCnt < 10) { // wait a little bit & continue
				nErrorCnt++;
				Sleep(100);
			} else {
				szError.Format(_T("Can't send data!\r\nError = %d.\r\n\r\nAlready sent %d packets / %d bytes.\r\nRest %d packets / %d bytes."),
					WSAGetLastError(), nSentCnt, nSendSize, m_nTCPPacketCnt - nSentCnt, m_nTCPDataSize - nSendSize);
				return false;
			}
		} else { // ok - go to next node
			nSentCnt++;
			nSendSize += node->m_nSize;
			node = node->m_pNext;
			nErrorCnt = 0;
		}
	}
	return true;
}

bool CSLVolaSplitterDlg::AddVolaData(const char *data, int size)
{
	TCPData *node = new TCPData(data, size);
	ATLASSERT(node != NULL);
	if (NULL == node)
		return false;
	ATLASSERT(node->m_pData != NULL);
	if (NULL == node->m_pData)
		return false;
	if (m_pTCPDataTail != NULL)  m_pTCPDataTail->m_pNext = node;
	else  m_pTCPDataHead = node;
	m_pTCPDataTail = node;
	m_nTCPDataSize += size;
	m_nTCPPacketCnt++;
	return true;
}

void CSLVolaSplitterDlg::UpdateClientInfo(int nIndex)
{
	CString str, tmp;
	if (INVALID_SOCKET != m_sClientSocket[nIndex])  str = _T("Run");
	else  str = _T("None");
	m_IPState[nIndex].GetWindowText(tmp);
	if (tmp != str)
		m_IPState[nIndex].SetWindowText(str);
}

void CSLVolaSplitterDlg::Test()
{
	//for (int i = 0; i < 999999; i++) {
	//	CStringA str;  str.Format("L01%06d", i);
	//	int ln = str.GetLength();
	//	AddVolaData((LPCSTR)str, ln + 1);
	//}
}
